import torch


class EuclideanDistance:

    @staticmethod
    def distance(fp1, fp2):
        '''
        Distance function between two fingerprints.
        '''
        return torch.linalg.norm(fp1 - fp2)


class BaseKernelType:
    '''
    Base class for all kernel types with common properties,
    attributes and methods.
    '''

    def __init__(self, weight=1.0, scale=1.0, metric=EuclideanDistance):
        # Currently, all the kernel types take weight and scale as parameters

        self.params = {'scale': scale, 'weight': weight}
        self.metric = metric

    @property
    def scale(self):        
        return self.params['scale']

    @property
    def weight(self):
        return self.params['weight']

    def update(self, params):
        '''
        Update the kernel function hyperparameters.
        '''

        self.params.update(params)


class SquaredExp(BaseKernelType):
    '''
    Squared Exponential kernel function
    '''

    def kernel_function(self, fp1, fp2):
        '''
        Kernel function between two atomic fingerprints 'fp1' and 'fp2'.
        '''
        
        return (self.weight**2 *
                torch.exp(-self.metric.distance(fp1, fp2)**2 / 2 / self.scale**2))
    
    def kernel_gradient(self, fp1, fp2, fp1_grad):
        '''
        Calculates the derivative of the kernel between
        fp1 and fp2 w.r.t. all coordinates in fp1.

        Chain rule:

                d k(x, x')    dk      d D(x, x')
                ---------- = ----  X  ----------
                   d xi       dD         d xi
        '''

        prefactor = - self.kernel_function(fp1, fp2) / self.scale**2
    
        return prefactor * torch.einsum('i,hi->h', fp1 - fp2, fp1_grad)

    def kernel_hessian(self, fp1, fp2, fp1_grad, fp2_grad):
        '''
        Kernel hessian w.r.t. atomic coordinates in both 'fp1' and 'fp2'

                    d^2 k(x, x')
                    ------------
                     dx_i dx'_j
        '''

        prefactor = self.kernel_function(fp1, fp2) / self.scale**2

        # [Nfeatures] * [Natoms * 3, Nfeatures] -> [Natoms * 3]
        DdD_dr1 = torch.einsum('i,hi->h', fp1 - fp2, fp1_grad)
        DdD_dr2 = torch.einsum('i,hi->h', fp2 - fp1, fp2_grad)

        # [Natoms * 3] * [Natoms * 3] -> [Natoms * 3, Natoms * 3]
        C0 = prefactor * (1 / self.scale**2) * torch.einsum('i,j->ij', DdD_dr1, DdD_dr2)
        
        # [Natoms * 3, Nfeatures] * [Natoms * 3, Nfeatures] -> [Natoms * 3, Natoms * 3]
        C1 = prefactor * torch.einsum('hi,ki->hk', fp1_grad, fp2_grad)

        return C0 + C1



import numpy as np
import copy
import time

import torch
from ase import io
from ase.atoms import Atoms
from ase.optimize import FIRE, MDMin, LBFGS, BFGS
from ase.parallel import parprint, parallel_function

try:
    from ase.mep import NEB, DyNEB
except ModuleNotFoundError:
    from ase.neb import NEB, DyNEB

from aenet_gpr.src import GPRCalculator
from aenet_gpr.util import ReferenceData
from aenet_gpr.tool import acquisition, prepare_neb_images, dump_observation, get_fmax
from aenet_gpr.inout.input_parameter import InputParameters


def min_cartesian_dist(img, train_images):

    best = np.inf
    for t in train_images:
        dt = img.positions.flatten() - t.positions.flatten()
        d2 = np.dot(dt, dt)
        if d2 < best:
            best = d2
    return best


class AIDNEB:

    def __init__(self, start, end, input_param: InputParameters, model_calculator=None, calculator=None,
                 interpolation='idpp', n_images=15, n_train_images=3, k=None, climbing=True, mic=False,
                 neb_method='improvedtangent',  # 'improvedtangent', 'aseneb'
                 remove_rotation_and_translation=False,
                 max_train_data=25, force_consistent=None,
                 max_train_data_strategy='nearest_observations',
                 trajectory='AIDNEB.traj',
                 use_previous_observations=False):

        """
        Artificial Intelligence-Driven Nudged Elastic Band (AID-NEB) algorithm.
        Optimize a NEB using a surrogate GPR model [1-3].
        Potential energies and forces at a given position are
        supplied to the model calculator to build a modelled PES in an
        active-learning fashion. This surrogate relies on NEB theory to
        optimize the images along the path in the predicted PES. Once the
        predicted NEB is optimized, the acquisition function collect a new
        observation based on the predicted energies and uncertainties of the
        optimized images. Gaussian Process Regression, aenet-gpr, is used to
        build the model as implemented in [4].

        [1] J. A. Garrido Torres, P. C. Jennings, M. H. Hansen,
        J. R. Boes, and T. Bligaard, Phys. Rev. Lett. 122, 156001 (2019).
        https://doi.org/10.1103/PhysRevLett.122.156001
        [2] O. Koistinen, F. B. Dagbjartsdóttir, V. Ásgeirsson, A. Vehtari,
        and H. Jónsson, J. Chem. Phys. 147, 152720 (2017).
        https://doi.org/10.1063/1.4986787
        [3] E. Garijo del Río, J. J. Mortensen, and K. W. Jacobsen,
        Phys. Rev. B 100, 104103 (2019).
        https://doi.org/10.1103/PhysRevB.100.104103
        [4] I. W. Yeu, A. Stuke, J. López-Zorrilla, J. M Stevenson, D. R Reichman, R. A Friesner,
        A. Urban, and N. Artrith, npj Computational Materials 11, 156 (2025).
        https://doi.org/10.1038/s41524-025-01651-0

        NEB Parameters
        --------------
        initial: Trajectory file (in ASE format) or Atoms object.
            Initial end-point of the NEB path.

        final: Trajectory file (in ASE format) or Atoms object.
            Final end-point of the NEB path.

        model_calculator: Model object.
            Model calculator to be used for predicting the potential energy
            surface. The default is None which uses a GP model with the Squared
            Exponential Kernel and other default parameters. See
            *ase.optimize.activelearning.gp.calculator* GPModel for default GP
            parameters.

        interpolation: string or Atoms list or Trajectory
            NEB interpolation.

            options:
                - 'linear' linear interpolation.
                - 'idpp'  image dependent pair potential interpolation.
                - Trajectory file (in ASE format) or list of Atoms.
                The user can also supply a manual interpolation by passing
                the name of the trajectory file  or a list of Atoms (ASE
                format) containing the interpolation images.

        mic: boolean
            Use mic=True to use the Minimum Image Convention and calculate the
            interpolation considering periodic boundary conditions.

        n_images: int or float
            Number of images of the path. Only applicable if 'linear' or
            'idpp' interpolation has been chosen.
            options:
                - int: Number of images describing the NEB. The number of
                images include the two (initial and final) end-points of the
                NEB path.
                - float: Spacing of the images along the NEB. The number of
                images is calculated as the length of the interpolated
                initial path divided by the spacing (Ang^-1).

        k: float or list
            Spring constant(s) in eV/Angstrom.

        neb_method: string
            NEB method as implemented in ASE. ('aseneb', 'improvedtangent'
            or 'eb'). See https://wiki.fysik.dtu.dk/ase/ase/neb.html.

        calculator: ASE calculator Object.
            ASE calculator.
            See https://wiki.fysik.dtu.dk/ase/ase/calculators/calculators.html

        force_consistent: boolean or None
            Use force-consistent energy calls (as opposed to the energy
            extrapolated to 0 K). By default (force_consistent=None) uses
            force-consistent energies if available in the calculator, but
            falls back on force_consistent=False if not.

        trajectory: string
            Filename to store the predicted NEB paths.
                Additional information:
                - Energy uncertain: The energy uncertainty in each image
                position can be accessed in image.info['uncertainty'].

        use_previous_observations: boolean
            If False. The optimization starts from scratch.
            A *trajectory_observations.traj* file is automatically generated
            in each step of the optimization, which contains the
            observations collected by the surrogate. If
            (a) *use_previous_observations* is True and (b) a previous
            *trajectory_observations.traj* file is found in the working
            directory: the algorithm will be use the previous observations
            to train the model with all the information collected in
            *trajectory_observations.traj*.

        max_train_data: int
            Number of observations that will effectively be included in the
            model. See also *max_data_strategy*.

        max_train_data_strategy: string
            Strategy to decide the observations that will be included in the
            model.

            options:
                'last_observations': selects the last observations collected by
                the surrogate.
                'lowest_energy': selects the lowest energy observations
                collected by the surrogate.
                'nearest_observations': selects the observations which
                positions are nearest to the positions of the Atoms to test.

            For instance, if *max_train_data* is set to 50 and
            *max_train_data_strategy* to 'lowest energy', the surrogate model
            will be built in each iteration with the 50 lowest energy
            observations collected so far.

        """

        # Convert Atoms and list of Atoms to trajectory files.
        if isinstance(start, Atoms):
            io.write('initial.traj', start)
            start = 'initial.traj'
        if isinstance(end, Atoms):
            io.write('final.traj', end)
            end = 'final.traj'

        # if isinstance(start, Atoms) and isinstance(end, Atoms):
        #     prepare_neb_images(start, end)
        #     start = '00_initial.traj'
        #     end = '01_final.traj'
        # else:
        #     raise ValueError("Both images must be Atoms object.")

        interp_path = None
        if interpolation != 'idpp' and interpolation != 'linear':
            interp_path = interpolation
        if isinstance(interp_path, list):
            io.write('initial_path.traj', interp_path)
            interp_path = 'initial_path.traj'

        # NEB parameters.
        self.input_param = input_param
        self.start = start
        self.end = end
        self.n_images = n_images
        self.n_train_images = n_train_images
        self.mic = mic
        self.rrt = remove_rotation_and_translation
        self.neb_method = neb_method
        self.spring = k
        self.climbing = climbing
        self.i_endpoint = io.read(self.start, '-1')
        self.e_endpoint = io.read(self.end, '-1')

        # GP calculator:
        self.model_calculator = model_calculator

        # Active Learning setup (Single-point calculations).
        self.step = 0
        self.function_calls = 0
        self.force_calls = 0
        self.ase_calc = calculator
        self.atoms = io.read(self.start, '-1')

        self.constraints = self.atoms.constraints
        self.force_consistent = force_consistent
        self.use_previous_observations = use_previous_observations
        self.trajectory = trajectory

        # Make sure that the initial and endpoints are near the interpolation.
        if self.mic:
            mic_initial = self.i_endpoint[:]
            mic_final = self.e_endpoint[:]
            mic_images = [mic_initial]
            for i in range(10000):
                mic_images += [mic_initial.copy()]
            mic_images += [mic_final]
            neb_mic = DyNEB(mic_images, climb=False, method=self.neb_method, remove_rotation_and_translation=self.rrt)
            neb_mic.interpolate(method='linear', mic=self.mic)
            self.i_endpoint.positions = mic_images[1].positions[:]
            self.e_endpoint.positions = mic_images[-2].positions[:]

        # Calculate the initial and final end-points (if necessary).
        if self.i_endpoint.calc is None:
            self.i_endpoint.calc = copy.deepcopy(self.ase_calc)
        if self.e_endpoint.calc is None:
            self.e_endpoint.calc = copy.deepcopy(self.ase_calc)
        self.i_endpoint.get_potential_energy(force_consistent=force_consistent)
        self.i_endpoint.get_forces()
        self.e_endpoint.get_potential_energy(force_consistent=force_consistent)
        self.e_endpoint.get_forces()

        if isinstance(self.i_endpoint, Atoms):
            io.write('initial.traj', self.i_endpoint)
        if isinstance(self.e_endpoint, Atoms):
            io.write('final.traj', self.e_endpoint)

        # Calculate the distance between the initial and final endpoints.
        self.d_start_end = np.sum((self.i_endpoint.positions.flatten() -
                                   self.e_endpoint.positions.flatten()) ** 2) ** 0.5

        self.rmax = 0.4 * self.d_start_end

        # A) Create images using interpolation if user does define a path.
        if interp_path is None:
            if isinstance(self.n_images, float):
                self.n_images = int(self.d_start_end / self.n_images)
            if self.n_images <= 3:
                self.n_images = 3
            self.images = make_neb(self)
            if self.spring is None:
                self.spring = 0.1 * (self.n_images - 1) / self.d_start_end
                # self.spring = 2. * np.sqrt(self.n_images - 1) / self.d_start_end ** 2

            neb_interpolation = DyNEB(self.images, climb=False, k=self.spring, method=self.neb_method, remove_rotation_and_translation=self.rrt)
            neb_interpolation.interpolate(method='linear', mic=self.mic)
            if interpolation == 'idpp':
                neb_interpolation = DyNEB(self.images, climb=True, k=self.spring, method=self.neb_method, remove_rotation_and_translation=self.rrt)
                neb_interpolation.interpolate(method='idpp', mic=self.mic)
                # neb_interpolation.idpp_interpolate(optimizer=FIRE, mic=self.mic)

        # B) Alternatively, the user can propose an initial path.
        if interp_path is not None:
            images_path = io.read(interp_path, ':')
            first_image = images_path[0].get_positions().reshape(-1)
            last_image = images_path[-1].get_positions().reshape(-1)

            is_pos = self.i_endpoint.get_positions().reshape(-1)
            fs_pos = self.e_endpoint.get_positions().reshape(-1)

            if not np.array_equal(first_image, is_pos):
                images_path.insert(0, self.i_endpoint)
            if not np.array_equal(last_image, fs_pos):
                images_path.append(self.e_endpoint)

            self.n_images = len(images_path)
            self.images = make_neb(self, images_interpolation=images_path)

        # Guess spring constant (k) if not defined by the user.
        if self.spring is None:
            self.spring = 0.1 * (self.n_images - 1) / self.d_start_end
            # self.spring = 2. * np.sqrt(self.n_images - 1) / self.d_start_end ** 2
        # Save initial interpolation.
        self.initial_interpolation = self.images[:]

        print()
        print('Distance between initial and final: ', self.d_start_end)
        # print(f"r_max (threshold to prevent over-relaxation when training data is sparse): {self.rmax:.4f}")
        print('spring_constant: ', self.spring)

    def run(self, fmax=0.05, unc_convergence=0.1, dt=0.1, ml_steps=150, optimizer="MDMin",
            max_unc_trheshold=1.0, check_ref_force=False):

        """
        Executing run will start the NEB optimization process.

        Parameters
        ----------
        fmax : float
            Convergence criteria (in eV/Angstrom).

        unc_convergence: float
            Maximum uncertainty for convergence (in eV). The algorithm's
            convergence criteria will not be satisfied if the uncertainty
            on any of the NEB images in the predicted path is above this
            threshold.

        dt : float
            dt parameter for MDMin.

        ml_steps: int
            Maximum number of steps for the NEB optimization on the
            modelled potential energy surface.

        max_unc_trheshold: float
            Safe control parameter. This parameter controls the degree of
            freedom of the NEB optimization in the modelled potential energy
            surface or the. If the uncertainty of the NEB lies above the
            'max_unc_trheshold' threshold the NEB won't be optimized and the image
            with maximum uncertainty is evaluated. This prevents exploring
            very uncertain regions which can lead to probe unrealistic
            structures.

        Returns
        -------
        Minimum Energy Path from the initial to the final states.

        """
        trajectory_main = self.trajectory.split('.')[0]
        trajectory_observations = trajectory_main + '_observations.traj'
        trajectory_candidates = trajectory_main + '_candidates.traj'

        # Start by saving the initial and final states.
        dump_observation(atoms=self.i_endpoint,
                         filename=trajectory_observations,
                         restart=self.use_previous_observations)
        self.use_previous_observations = True  # Switch on active learning.
        dump_observation(atoms=self.e_endpoint,
                         filename=trajectory_observations,
                         restart=self.use_previous_observations)

        print(f"[INFO] Number of initial training data: {self.n_train_images} including initial and final")
        n_to_add = min(self.n_images - 2, max(1, int(self.d_start_end / 3.0)))
        if (n_to_add + 2) > self.n_train_images:
            print(f"[INFO] Distance between initial and final (ΔR = {self.d_start_end} Å) is too large")
            print(f"[INFO] Use {n_to_add + 2} initial training data instead of {self.n_train_images}")
            self.n_train_images = n_to_add + 2

        train_images = io.read(trajectory_observations, ':')
        if len(train_images) == 2:
            # middle = int(self.n_images * (2. / 3.))
            # e_is = self.i_endpoint.get_potential_energy()
            # e_fs = self.e_endpoint.get_potential_energy()
            #
            # if e_is > e_fs:
            #     middle = int(self.n_images * (1. / 3.))
            n_to_add = self.n_train_images - 2
            middle_indices = [int(round(i * (self.n_images - 1) / (n_to_add + 1))) for i in range(1, n_to_add + 1)]

            for middle_idx in middle_indices:
                print(f"[INFO] Adding training image at NEB index {middle_idx}")

                self.atoms.positions = self.images[middle_idx].get_positions()
                self.atoms.calc = self.ase_calc

                # Reference calculation
                self.atoms.get_potential_energy(force_consistent=self.force_consistent)
                self.atoms.get_forces()
                dump_observation(atoms=self.atoms, method='neb',
                                 filename=trajectory_observations,
                                 restart=self.use_previous_observations)
                self.function_calls += 1
                self.force_calls += 1
                self.step += 1

        weight_update = self.input_param.weight
        scale_update = self.input_param.scale

        self.rmin = 0.1
        self.max_unc_hist = []
        self.max_unc_hist_after = []

        violated_index = 0
        set_update_step = False
        update_step = 10

        while True:

            # 0. Start from initial interpolation every 50 steps.
            # if self.step % 50 == 0:
            #     parprint('Starting from initial interpolation...')
            #     self.images = copy.deepcopy(self.initial_interpolation)

            # 1. Collect observations.
            # This serves to use_previous_observations from a previous (and/or parallel) runs.
            train_images = io.read(trajectory_observations, ':')

            train_data = ReferenceData(structure_files=train_images,
                                       file_format='ase',
                                       device=self.input_param.device,
                                       descriptor=self.input_param.descriptor,
                                       data_type=self.input_param.data_type,
                                       data_process=self.input_param.data_process,
                                       soap_param=self.input_param.soap_param,
                                       standardization=False,
                                       mask_constraints=self.input_param.mask_constraints)

            if train_data.standardization:
                train_data.standardize_energy_force(train_data.energy)

            # 2. Prepare a calculator.
            print('Training data size: ', len(train_images))
            print('Descriptor: ', self.input_param.descriptor)

            if violated_index is None and not set_update_step:
                update_step = self.step
                set_update_step = True

            if self.step >= update_step:
                update_step *= 2
                self.input_param.fit_weight = True
                self.input_param.fit_scale = True

                if self.input_param.filter:
                    train_data.filter_similar_data(threshold=self.rmin)
                    print('Actual training data size (after removing similar data): ', len(train_data.images))

                train_data.config_calculator(kerneltype='sqexp',
                                             scale=scale_update,
                                             weight=weight_update,
                                             noise=self.input_param.noise,
                                             noisefactor=self.input_param.noisefactor,
                                             use_forces=self.input_param.use_forces,
                                             sparse=self.input_param.sparse,
                                             sparse_derivative=self.input_param.sparse_derivative,
                                             autograd=self.input_param.autograd,
                                             train_batch_size=self.input_param.train_batch_size,
                                             eval_batch_size=self.input_param.eval_batch_size,
                                             fit_weight=self.input_param.fit_weight,
                                             fit_scale=self.input_param.fit_scale)

            else:
                self.input_param.fit_weight = False
                self.input_param.fit_scale = False

                if self.input_param.filter:
                    train_data.filter_similar_data(threshold=self.rmin)
                    print('Actual training data size (after removing similar data): ', len(train_data.images))

                train_data.config_calculator(kerneltype='sqexp',
                                             scale=scale_update,
                                             weight=weight_update,
                                             noise=self.input_param.noise,
                                             noisefactor=self.input_param.noisefactor,
                                             use_forces=self.input_param.use_forces,
                                             sparse=self.input_param.sparse,
                                             sparse_derivative=self.input_param.sparse_derivative,
                                             autograd=self.input_param.autograd,
                                             train_batch_size=self.input_param.train_batch_size,
                                             eval_batch_size=self.input_param.eval_batch_size,
                                             fit_weight=self.input_param.fit_weight,
                                             fit_scale=self.input_param.fit_scale)

            print(f"r_min (threshold to avoid adding nearly-duplicate training data): {self.rmin:.2f}")
            print('GPR model hyperparameters: ', train_data.calculator.hyper_params)

            self.model_calculator = GPRCalculator(calculator=train_data.calculator, train_data=train_data)
            weight_update = train_data.calculator.weight.clone().detach().item()
            scale_update = train_data.calculator.scale.clone().detach().item()

            # Detach calculator from the prev. optimized images (speed up).
            # for i in self.images:
            #     i.calc = None
            # # Train only one process.
            # calc.update_train_data(train_images, test_images=self.images)
            # Attach the calculator (already trained) to each image.
            for i in self.images:
                i.calc = copy.deepcopy(self.model_calculator)

            # 3. Optimize the NEB in the predicted PES.
            # Get path uncertainty for deciding whether NEB or CI-NEB.
            predictions = get_neb_predictions(self.images)
            neb_pred_energy = predictions['energy']
            # neb_pred_forces = predictions['forces']
            neb_pred_uncertainty = predictions['uncertainty']

            # Climbing image NEB mode is risky when the model is trained with a few data points.
            # Switch on climbing image only when the uncertainty of the NEB the force of the climbing image are low.
            climbing_neb = False
            if self.climbing:
                max_unc = np.max(neb_pred_uncertainty)
                ok_unc = max_unc <= unc_convergence

                ci_idx = np.argmax(neb_pred_energy)  # default climbing candidate
                ci_unc = neb_pred_uncertainty[ci_idx]
                # ci_force = np.linalg.norm(neb_pred_forces[ci_idx])
                ci_ok = (ci_unc <= 0.1)  # and (ci_force <= 1.0)

                self.max_unc_hist.append(max_unc)
                W = 7
                ok_stall_unc = False
                if len(self.max_unc_hist) > W:
                    dec_unc = np.diff(np.array(self.max_unc_hist[-W:]))
                    ok_stall_unc = np.all(np.abs(dec_unc) <= 0.02)

                if (ok_unc or ok_stall_unc) and ci_ok:
                    parprint(f"Climbing image is now activated at image {ci_idx}.")
                    climbing_neb = True
            else:
                pass

            ml_neb = DyNEB(self.images, climb=climbing_neb, method=self.neb_method, k=self.spring)
            # FIRE, MDMin, LBFGS, BFGS
            if optimizer.lower() == 'mdmin':
                neb_opt = MDMin(ml_neb, dt=dt, trajectory="gpr_neb.traj")
            elif optimizer.lower() == 'lbfgs':
                neb_opt = LBFGS(ml_neb, trajectory="gpr_neb.traj")
            elif optimizer.lower() == 'bfgs':
                neb_opt = BFGS(ml_neb, trajectory="gpr_neb.traj")
            else:
                neb_opt = FIRE(ml_neb, dt=dt, trajectory="gpr_neb.traj")

            # Safe check to optimize the images.
            nim = len(self.images) - 2
            nat = len(self.images[0])
            if np.max(neb_pred_uncertainty) <= max_unc_trheshold:
                neb_opt.run(fmax=fmax * 1.0, steps=ml_steps)

                # previous position snapshot (for rollback)
                # ok_forces = False
                # prev_positions = [im.get_positions().copy() for im in self.images]
                #
                # for step in range(ml_steps):
                #     neb_opt.run(fmax=fmax * 1.0, steps=1)
                #
                #     #
                #     violated_index = None
                #     for i in range(1, len(self.images) - 1):
                #         d2 = min_cartesian_dist(self.images[i], train_images)
                #         if d2 > (self.rmax ** 2):
                #             violated_index = i
                #             print(f"Relaxation stopped early: image {violated_index} lies farther than r_max from all training data")
                #             break
                #
                #     if violated_index is not None:
                #         for im, pos in zip(self.images, prev_positions):
                #             im.set_positions(pos)
                #
                #         break
                #     else:
                #         prev_positions = [im.get_positions().copy() for im in self.images]
                #
                #     if neb_opt.converged(ml_neb.get_forces().ravel()):
                #         ok_forces = True
                #         break

                F = ml_neb.get_forces()  # (n_mobile*nat*3) flat
                F = F.reshape(nim, nat, 3)
                max_f_image = np.sqrt((F ** 2).sum(-1)).max().item()

            else:
                parprint("The uncertainty of the NEB lies above the max_unc threshold (1.0).")
                parprint("NEB won't be optimized and the image with maximum uncertainty is just evaluated and added")
                F = np.array([im.get_forces() for im in self.images[1:-1]])
                max_f_image = np.sqrt((F ** 2).sum(-1)).max().item()

            predictions = get_neb_predictions(self.images)
            neb_pred_energy = predictions['energy']
            # neb_pred_forces = predictions['forces']
            neb_pred_uncertainty = predictions['uncertainty']

            # 5. Print output.
            max_e = np.max(neb_pred_energy)
            max_unc = np.max(neb_pred_uncertainty)
            # max_e_ind = np.argsort(neb_pred_energy)[-1]

            # Calculator of train_images is reference, while Calculator of self.images is GP
            if check_ref_force:
                max_f = get_fmax(train_images[-1])
            else:
                max_f = max_f_image

            pbf = max_e - self.i_endpoint.get_potential_energy(force_consistent=self.force_consistent)
            pbb = max_e - self.e_endpoint.get_potential_energy(force_consistent=self.force_consistent)

            msg = "--------------------------------------------------------"
            parprint(msg)
            parprint('Step:', self.step)
            parprint('Time:', time.strftime("%m/%d/%Y, %H:%M:%S", time.localtime()))
            parprint('Predicted barrier (-->):', pbf)
            parprint('Predicted barrier (<--):', pbb)
            parprint('Number of images:', len(self.images))
            parprint('Max. uncertainty:', max_unc)
            parprint("Max. force:", max_f)
            msg = "--------------------------------------------------------\n"
            parprint(msg)

            # 6. Check convergence.
            ok_unc = max_unc <= unc_convergence

            self.max_unc_hist_after.append(max_unc)
            ok_stall_unc = False
            if len(self.max_unc_hist_after) > W:
                dec_unc = np.diff(np.array(self.max_unc_hist_after[-W:]))
                ok_stall_unc = np.all(np.abs(dec_unc) <= 0.02)

            # Max.forces and NEB images uncertainty must be below *fmax* and *unc_convergence* thresholds.
            if len(train_images) > 2 and max_f <= fmax and (ok_unc or ok_stall_unc) and max_unc < unc_convergence * 5 and (climbing_neb or not self.climbing):
                parprint('A saddle point was found.')

                # if np.max(neb_pred_uncertainty[1:-1]) < unc_convergence:
                io.write(self.trajectory, self.images)
                parprint('Uncertainty of the images below threshold.')
                parprint('NEB converged.')
                parprint('The NEB path can be found in:', self.trajectory)
                msg = "Visualize the last path using 'ase gui "
                msg += self.trajectory
                parprint(msg)
                break

            # Set the path to previous iterations if barrier is higher than 10 eV
            # if pbf > 10.0:
            #     print(f"[INFO] Current energy barrier (ΔE = {pbf} eV) is too large, meaning largely deviated path")
            #     print('Reset the images to the previous path...')
            #     self.images = copy.deepcopy(io.read("gpr_neb.traj", f":{self.n_images}"))
            #     for i in self.images:
            #         i.calc = copy.deepcopy(self.model_calculator)

            # 7. Select next point to train (acquisition function):
            # Candidates are the optimized NEB images in the predicted PES.
            candidates = copy.deepcopy(self.images)[1:-1]

            if np.max(neb_pred_uncertainty) > unc_convergence:
                sorted_candidates = acquisition(train_images=train_images,
                                                candidates=candidates,
                                                mode='uncertainty',
                                                objective='max')
            else:
                if self.step % 5 == 0:
                    sorted_candidates = acquisition(train_images=train_images,
                                                    candidates=candidates,
                                                    mode='fmax',
                                                    objective='min')
                else:
                    sorted_candidates = acquisition(train_images=train_images,
                                                    candidates=candidates,
                                                    mode='ucb',
                                                    objective='max')

            # Select the best candidate.
            accepted = False
            fp_train = train_data.generate_cartesian(train_data.images)
            N = fp_train.shape[0]

            while not accepted:
                sorted_candidates_tmp = copy.deepcopy(sorted_candidates)
                found_candidate = False

                while sorted_candidates_tmp:
                    best_candidate = sorted_candidates_tmp.pop(0)
                    fp_candidate = train_data.generate_cartesian_per_data(best_candidate).flatten()

                    for i in range(N):
                        xi = fp_train[i].flatten()
                        dist = torch.linalg.norm(xi - fp_candidate)

                        if dist < self.rmin:
                            print(f"Candidate rejected: too close to train data {i} (dist={dist:.4f} < r_min={self.rmin:.3f})")
                            break
                    else:
                        accepted = True
                        found_candidate = True
                        chosen_candidate = best_candidate
                        break

                if not found_candidate and not accepted:
                    self.rmin -= 0.01
                    print(f"No candidate accepted. Decrease r_min to {self.rmin:.3f} and retry...")

            # Save the other candidates for multi-task optimization.
            if sorted_candidates:
                io.write(trajectory_candidates, sorted_candidates)

            # 8. Evaluate the target function and save it in *observations*.
            self.atoms.positions = chosen_candidate.get_positions()
            self.atoms.calc = self.ase_calc
            self.atoms.get_potential_energy(force_consistent=self.force_consistent)
            self.atoms.get_forces()
            dump_observation(atoms=self.atoms,
                             filename=trajectory_observations,
                             restart=self.use_previous_observations)
            self.function_calls += 1
            self.force_calls += 1
            self.step += 1
        print_cite_neb()


@parallel_function
def make_neb(self, images_interpolation=None):
    """
    Creates a NEB from a set of images.
    """
    imgs = [self.i_endpoint[:]]
    for i in range(1, self.n_images - 1):
        image = self.i_endpoint[:]
        if images_interpolation is not None:
            image.set_positions(images_interpolation[i].get_positions())
        image.set_constraint(self.constraints)
        imgs.append(image)
    imgs.append(self.e_endpoint[:])
    return imgs


@parallel_function
def get_neb_predictions(images):
    neb_pred_energy = []
    neb_pred_forces = []
    neb_pred_unc = []

    for i in images:
        neb_pred_energy.append(i.get_potential_energy())
        neb_pred_forces.append(i.get_forces())
        unc = i.calc.results['uncertainty']
        neb_pred_unc.append(unc)

    neb_pred_unc[0] = 0.0
    neb_pred_unc[-1] = 0.0
    predictions = {'energy': neb_pred_energy, 'forces': neb_pred_forces, 'uncertainty': neb_pred_unc}

    return predictions


@parallel_function
def print_cite_neb():
    msg = "\n" + "-" * 79 + "\n"
    msg += "You are using GPR-accelerated NEB. Please cite: \n"
    msg += "[1] J. A. Garrido Torres, P. C. Jennings, M. H. Hansen, "
    msg += "J. R. Boes, and T. Bligaard, Phys. Rev. Lett. 122, 156001 (2019). "
    msg += "https://doi.org/10.1103/PhysRevLett.122.156001 \n"
    msg += "[2] O. Koistinen, F. B. Dagbjartsdóttir, V. Ásgeirsson, A. Vehtari,"
    msg += " and H. Jónsson, J. Chem. Phys. 147, 152720 (2017). "
    msg += "https://doi.org/10.1063/1.4986787 \n"
    msg += "[3] E. Garijo del Río, J. J. Mortensen, and K. W. Jacobsen, "
    msg += "Phys. Rev. B 100, 104103 (2019). "
    msg += "https://doi.org/10.1103/PhysRevB.100.104103. \n"
    msg += "[4] I. W. Yeu, A. Stuke, J. López-Zorrilla, J. M Stevenson, D. R Reichman, R. A Friesner, "
    msg += "A. Urban, and N. Artrith, npj Computational Materials 11, 156 (2025)."
    msg += "https://doi.org/10.1038/s41524-025-01651-0. \n"
    msg += "-" * 79 + '\n'
    parprint(msg)



import torch
import torch.nn as nn
import numpy as np

from aenet_gpr.src.prior import ConstantPrior
from aenet_gpr.src.pytorch_kernel_internal import FPKernelInternal
from aenet_gpr.util.prepare_data import get_N_batch, get_batch_indexes_N_batch


class GaussianProcessInternal(nn.Module):
    '''
    Gaussian Process Regression
    Parameters:

    prior: Defaults to ConstantPrior with zero as constant

    kernel: Defaults to the Squared Exponential kernel with derivatives
    '''

    def __init__(self, hp=None, prior=None, prior_update=True, kerneltype='sqexp',
                 scale_b=0.4, scale_a=0.1, scale_d=0.2, weight=1.0, noise=1e-6, noisefactor=0.5,
                 use_forces=True, images=None, function=None, derivative=None,
                 sparse=None, sparse_derivative=None, autograd=False,
                 train_batch_size=25, eval_batch_size=25,
                 data_type='float64', device='cpu', soap_param=None, descriptor='internal',
                 atoms_mask=None, c_table=None):
        super().__init__()

        if data_type == 'float32':
            self.data_type = 'float32'
            self.torch_data_type = torch.float32
        else:
            self.data_type = 'float64'
            self.torch_data_type = torch.float64

        self.device = device
        self.soap_param = soap_param
        self.descriptor = descriptor
        self.kerneltype = kerneltype

        self.scale = {'bond': torch.tensor(scale_b, dtype=self.torch_data_type, device=self.device),
                      'angle': torch.tensor(scale_a, dtype=self.torch_data_type, device=self.device),
                      'dihedral': torch.tensor(scale_d, dtype=self.torch_data_type, device=self.device), }

        self.weight = torch.tensor(weight, dtype=self.torch_data_type, device=self.device)

        self.noise = torch.tensor(noise, dtype=self.torch_data_type, device=self.device)
        self.noisefactor = torch.tensor(noisefactor, dtype=self.torch_data_type, device=self.device)

        self.use_forces = use_forces
        self.images = images
        self.Ntrain = len(self.images)
        self.species = self.images[0].get_chemical_symbols()
        self.pbc = np.all(self.images[0].get_pbc())

        self.Natom = len(self.species)
        self.atoms_mask = atoms_mask
        self.c_table = c_table

        if self.atoms_mask is not None:
            self.Nmask = self.atoms_mask.shape[0]
        else:
            self.Nmask = 3 * self.Natom

        self.Y = function  # Y = [Ntrain]
        self.dY = derivative  # dY = [Ntrain, Natom, 3]
        self.model_vector = torch.empty((self.Ntrain * (1 + self.Nmask),), dtype=self.torch_data_type, device=self.device)

        if prior is None:
            self.prior = ConstantPrior(0.0, dtype=self.torch_data_type, device=self.device, atoms_mask=self.atoms_mask)
        else:
            self.prior = prior
        self.prior_update = prior_update

        self.sparse = sparse
        self.train_batch_size = train_batch_size
        self.eval_batch_size = eval_batch_size

        if sparse is not None:
            self.sX = sparse  # sX = [Nsparse, Nscenter, Nfeature]
            self.sparse = True

            if sparse_derivative is not None:
                self.sdX = sparse_derivative  # sdX = [Nsparse, Nscenter, Natom, 3, Nfeature]
            else:
                self.sdX = None

        else:
            self.sX = None
            self.sparse = False

        if self.use_forces:
            self.kernel = FPKernelInternal(species=self.species,
                                           pbc=self.pbc,
                                           Natom=self.Natom,
                                           kerneltype=self.kerneltype,
                                           data_type=self.data_type,
                                           soap_param=self.soap_param,
                                           descriptor=self.descriptor,
                                           device=self.device,
                                           atoms_mask=self.atoms_mask,
                                           c_table=self.c_table)

        hyper_params = dict(kerneltype=self.kerneltype,
                            scale=self.scale,
                            weight=self.weight,
                            noise=self.noise,
                            noisefactor=self.noisefactor,
                            prior=self.prior.constant)

        self.hyper_params = hyper_params
        self.kernel.set_params(self.hyper_params)

        if self.Y is not None:
            if self.dY is not None:
                # [Ntrain] -> [Ntrain, 1]
                # Y_reshaped = self.Y.flatten().unsqueeze(1)
                Y_reshaped = self.Y.contiguous().view(-1, 1)

                # [Ntrain, Natom, 3] -> [Ntrain * 3 * Natom, 1]
                # dY_reshaped = self.dY.flatten().unsqueeze(1)
                if self.atoms_mask is not None:
                    dY_reshaped = self.dY.contiguous().view(self.dY.shape[0], -1)[:, self.atoms_mask]
                    dY_reshaped = dY_reshaped.view(-1, 1)
                else:
                    dY_reshaped = self.dY.contiguous().view(-1, 1)

                # [Ntrain * (1 + 3 * Natom), 1]
                # [[e1, e2, ..., eN, f11x, f11y, f11z, f12x, f12y, ..., fNzNz]],
                self.YdY = torch.cat((Y_reshaped, dY_reshaped), dim=0)

                del Y_reshaped, dY_reshaped

            else:
                self.YdY = self.Y.flatten().unsqueeze(1)  # no dY [e1, e2, ..., eN]
        else:
            self.YdY = None

    def train_model(self):
        # covariance matrix between the training points X
        self.K_XX_L = self.kernel.kernel_matrix_batch(images=self.images, batch_size=self.train_batch_size)

        a = torch.full((self.Ntrain, 1), self.hyper_params['noise'] * self.hyper_params['noisefactor'],
                       dtype=self.torch_data_type, device=self.device)
        noise_val = self.hyper_params['noise']
        b = noise_val.expand(self.Ntrain, self.Nmask)

        # reg = torch.diag(torch.cat((a, b), 1).flatten() ** 2)
        diagonal_values = torch.cat((a, b), 1).flatten() ** 2

        self.K_XX_L.diagonal().add_(diagonal_values)

        try:
            self.K_XX_L = torch.linalg.cholesky(self.K_XX_L, upper=False)
        except torch.linalg.LinAlgError:
            with torch.no_grad():
                # Diagonal sum (trace)
                diag_sum = torch.sum(torch.diag(self.K_XX_L))

                # epsilon value
                eps = torch.finfo(self.torch_data_type).eps

                # scaling factor
                scaling_factor = 1 / (1 / (4.0 * eps) - 1)

                # adjust K_XX
                adjustment = diag_sum * scaling_factor * torch.ones(self.K_XX_L.shape[0],
                                                                    dtype=self.torch_data_type,
                                                                    device=self.device)
                self.K_XX_L.diagonal().add_(adjustment)

                # Step 1: Cholesky decomposition for K_XX after adjusting
                self.K_XX_L = torch.linalg.cholesky(self.K_XX_L, upper=False)

        if self.prior_update:
            self.prior.update(self.images, self.YdY, self.K_XX_L, self.use_forces)
            self.hyper_params.update(dict(prior=self.prior.constant))
            self.kernel.set_params(self.hyper_params)

        _prior_array = self.prior.potential_batch(self.images, self.use_forces)
        self.model_vector = torch.cholesky_solve(self.YdY.contiguous().view(-1, 1) - _prior_array.view(-1, 1),
                                                 self.K_XX_L, upper=False)

        return

    def forward(self, eval_images, get_variance=False):

        Ntest = len(eval_images)
        eval_x_N_batch = get_N_batch(Ntest, self.eval_batch_size)
        eval_x_indexes = get_batch_indexes_N_batch(Ntest, eval_x_N_batch)

        if not get_variance:
            if self.use_forces:
                # E_hat = pred[0:x.shape[0]]
                # F_hat = pred[x.shape[0]:].reshape(x.shape[0], self.Natom, -1)
                E_hat = torch.empty((Ntest,), dtype=self.torch_data_type)
                F_hat = torch.zeros((Ntest, self.Natom * 3), dtype=self.torch_data_type)

                for i in range(0, eval_x_N_batch):
                    pred, kernel = self.eval_data_batch(eval_images=eval_images[eval_x_indexes[i][0]:eval_x_indexes[i][1]])

                    data_per_batch = eval_x_indexes[i][1] - eval_x_indexes[i][0]
                    E_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1]] = pred[0:data_per_batch]
                    if self.atoms_mask is not None:
                        F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], self.atoms_mask] = pred[data_per_batch:].view(data_per_batch, -1)
                    else:
                        F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = pred[data_per_batch:].view(data_per_batch, -1)

                return E_hat, F_hat.view((Ntest, self.Natom, 3)), None

            else:
                pass

        else:
            if self.use_forces:
                E_hat = torch.empty((Ntest,), dtype=self.torch_data_type)
                F_hat = torch.zeros((Ntest, self.Natom * 3), dtype=self.torch_data_type)
                uncertainty = torch.empty((Ntest,), dtype=self.torch_data_type)

                for i in range(0, eval_x_N_batch):
                    pred, kernel = self.eval_data_batch(eval_images=eval_images[eval_x_indexes[i][0]:eval_x_indexes[i][1]])
                    var = self.eval_variance_batch(get_variance=get_variance,
                                                   eval_images=eval_images[eval_x_indexes[i][0]:eval_x_indexes[i][1]],
                                                   k=kernel)

                    data_per_batch = eval_x_indexes[i][1] - eval_x_indexes[i][0]
                    E_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1]] = pred[0:data_per_batch]
                    if self.atoms_mask is not None:
                        F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], self.atoms_mask] = pred[data_per_batch:].view(data_per_batch, -1)
                    else:
                        F_hat[eval_x_indexes[i][0]:eval_x_indexes[i][1], :] = pred[data_per_batch:].view(data_per_batch, -1)

                    uncertainty[eval_x_indexes[i][0]:eval_x_indexes[i][1]] = torch.sqrt(torch.diagonal(var)[0:data_per_batch]) / self.weight

                return E_hat, F_hat.view((Ntest, self.Natom, 3)), uncertainty

            else:
                pass

    def eval_data_batch(self, eval_images):

        kernel = self.kernel.kernel_vector_batch(eval_images=eval_images,
                                                 train_images=self.images,
                                                 batch_size=self.eval_batch_size)

        pred = torch.matmul(kernel, self.model_vector.view(-1)) + self.prior.potential_batch(eval_images, self.use_forces)

        return pred, kernel

    def eval_variance_batch(self, get_variance, eval_images, k):
        """

        variance.shape  # [Ntest * (1 + 3 * Natom), Ntest * (1 + 3 * Natom)]
        self.K_XX_L.shape  # [Ntrain * (1 + 3 * Natom), Ntrain * (1 + 3 * Natom)]
        k.T.clone().shape  # [Ntrain * (1 + 3 * Natom), Ntest * (1 + 3 * Natom)]
        self.Ck.shape  # [Ntrain * (1 + 3 * Natom), Ntest * (1 + 3 * Natom)]
        covariance.shape  # [Ntest * (1 + 3 * Natom), Ntest * (1 + 3 * Natom)]
        """
        # var = None
        if get_variance:
            covariance = torch.matmul(k, torch.cholesky_solve(k.T.clone(), self.K_XX_L, upper=False))

            return self.kernel.kernel_matrix_batch(images=eval_images, batch_size=self.eval_batch_size) - covariance

        else:
            return None

    def save_data(self, file="calc_dict.pt"):
        """
        self.data_type
        self.torch_data_type

        self.device = device
        self.noise
        self.noisefactor
        self.scale
        self.weight
        self.use_forces
        self.sparse

        (self.train_batch_size)
        (self.eval_batch_size)

        self.images
        self.Y
        self.dY
        self.YdY

        self.Ntrain
        self.Natom

        self.K_XX_L
        self.model_vector
        """

        state = {
            'kerneltype': self.kerneltype,
            'noise': self.noise,
            'noisefactor': self.noisefactor,
            'scale': self.scale,
            'weight': self.weight,
            'use_forces': self.use_forces,
            'sparse': self.sparse,
            'train_batch_size': self.train_batch_size,
            'eval_batch_size': self.eval_batch_size,
            'Y': self.Y,
            'dY': self.dY,
            'YdY': self.YdY,
            'K_XX_L': self.K_XX_L,
            'model_vector': self.model_vector,
        }
        torch.save(state, file)

    def load_data(self, file="calc_dict.pt"):
        state = torch.load(file)

        self.kerneltype = state.get('kerneltype')
        self.noise = state.get('noise')
        self.noisefactor = state.get('noisefactor')
        self.scale = state.get('scale')
        self.weight = state.get('weight')

        self.use_forces = state.get('use_forces')
        self.sparse = state.get('sparse')

        self.train_batch_size = state.get('train_batch_size')
        self.eval_batch_size = state.get('eval_batch_size')

        self.Y = state.get('Y')
        self.dY = state.get('dY')
        self.YdY = state.get('YdY')

        self.K_XX_L = state.get('K_XX_L')
        self.model_vector = state.get('model_vector')

        hyper_params = dict(kerneltype=self.kerneltype,
                            scale=self.scale,
                            weight=self.weight,
                            noise=self.noise,
                            noisefactor=self.noisefactor)

        self.hyper_params = hyper_params
        self.kernel.set_params(self.hyper_params)



from .acquisition import acquisition
from .trainingset import dump_observation, get_fmax, dump2list, TrainingSet
from .ase_tool import prepare_neb_images, align_molecule_centers, pbc_align, pbc_group, pbc_wrap, csv2list
from .aidneb import AIDNEB

__all__ = ["acquisition", "acquisition",
           "trainingset", "dump_observation", "get_fmax", "dump2list", "TrainingSet",
           "ase_tool", "prepare_neb_images", "align_molecule_centers", "pbc_align", "pbc_group", "pbc_wrap", "csv2list",
           "aidneb", "AIDNEB", ]



import numpy as np
import os
import warnings

from ase.parallel import parallel_function
from ase import io
from ase.io.trajectory import TrajectoryWriter
from ase.calculators.singlepoint import SinglePointCalculator, all_properties
from ase.calculators.calculator import PropertyNotImplementedError


@parallel_function
def dump_observation(atoms, filename, restart, method='-'):
    """
    Saves a trajectory file containing the atoms observations.

    Parameters
    ----------
    atoms: object
        Atoms object to be appended to previous observations.
    filename: string
        Name of the trajectory file to save the observations.
    restart: boolean
        Append mode (true or false).
    method: string
        Label with the optimizer name to be appended in atoms.info['method'].
     """
    atoms.info['method'] = method

    if restart:
        if os.path.exists(filename):
            prev_atoms = io.read(filename, ':')  # Active learning.
            if atoms not in prev_atoms:  # Avoid duplicates.
                # Update observations.
                trj = TrajectoryWriter(atoms=atoms, filename=filename,
                                       mode='a')
                trj.write()
        else:
            io.write(filename=filename, images=atoms, append=False)
    if not restart:
        io.write(filename=filename, images=atoms, append=False)


@parallel_function
def get_fmax(atoms):
    """
    Returns fmax for a given atoms structure.
    """
    forces = atoms.get_forces()
    return np.sqrt((forces**2).sum(axis=1).max())


def dump2list(atoms, destination, method):
    """
    Append the atoms object together with its calculator into
    a list. This is some sort of a copy and paste method.
    """

    # Copy the calculator
    calc = atoms.calc

    if calc is not None:
        results = {}
        implemented_properties = []

        for prop in all_properties:
            try:
                x = calc.get_property(prop, atoms,
                                      allow_calculation=False)

            except PropertyNotImplementedError:
                x = None
            results[prop] = x
            implemented_properties.append(prop)
        calculator = SinglePointCalculator(atoms, **results)
        calculator.name = calc.name
        calculator.implemented_properties = implemented_properties

        if hasattr(calc, 'parameters'):
            calculator.parameters.update(calc.parameters)

    # Copy the atoms object. The new object has no calculator
    atoms = atoms.copy()
    atoms.info['method'] = method
    atoms.calc = calculator

    destination.append(atoms)


class TrainingSet:

    """
    Training set class to store calculations as they are run
    by the different AID methods.

    It allows to save stuff to a trajectory file or to keep
    observations in memory as a list. In the future it would
    be nice to have them also stored to a database

    TODO: Add a db mode.

    Parameters:
    -----------
    destination: str or list
        This is where the training set is going to be stored
        options:
            trajectory file name: atoms are stored in a traj file.
                This is, written to disk.
            list: atoms are kept in memory, kept in a list.
    use_previous_observations: bool
        This is some sort of restart for the dataset,
        whether you would like to use the other stuff in your
        destination or not.
    atoms: Atoms object
        Atoms object to be written or dumpen into the training set
    """

    def __init__(self, destination, use_previous_observations, atoms=None):

        self.atoms = atoms
        if type(destination) is list:
            if not use_previous_observations:
                if len(destination) > 0:
                    warning = ("use_previous_observations == False together "
                               "with a non empty list as destintion deletes "
                               "the content of the list. If this is an "
                               "unwanted behaviour, consider passing an empty "
                               "list.")
                    warnings.warn(warning)

                self.destination = []
            else:
                self.destination = destination
            self.mode = 'list'
        elif type(destination) is str:
            if not destination.endswith('.traj'):
                msg = "*destination* should be a trajectory file"
                raise NotImplementedError(msg)
            else:
                self.destination = destination
                self.use_prev_obs = use_previous_observations
                self.mode = 'traj'
        else:
            msg = "*destination* should be a file or a list"
            raise NotImplementedError(msg)

    def dump(self, atoms=None, method='-'):
        """
        Saves the atoms observations into the training set.

        Parameters
        ----------
        atoms: object
            Atoms object to be appended to previous observations.
        method: string
            Label with the optimizer name to be appended in
            atoms.info['method'].
         """

        if atoms is None:
            atoms = self.atoms

        if self.mode == 'list':
            dump2list(atoms, destination=self.destination, method=method)

        elif self.mode == 'traj':
            if not self.use_prev_obs:
                dump_observation(atoms, filename=self.destination,
                                 method=method, restart=False)
                self.use_prev_obs = True
            else:
                dump_observation(atoms, filename=self.destination,
                                 method=method, restart=True)
        else:
            raise NotImplementedError()

    def write(self, atoms=None, method='-'):
        """
        Copy of the dump method,
        so that the training set can become an observer.
        """
        self.dump(atoms, method)

    def load_set(self):
        """
        Loads the training set

        Returns:
        --------
        A list of ase.Atoms objects
        """
        if self.mode == 'list':
            return self.destination.copy()
        elif self.mode == 'traj':
            return io.read(self.destination, ':')
        else:
            raise NotImplementedError()

    def load_last(self):
        """
        Last entry written to the training set

        Returns:
        --------
        An atoms object
        """
        if self.mode == 'list':
            return self.destination[-1]
        elif self.mode == 'traj':
            return io.read(self.destination, -1)
        else:
            raise NotImplementedError()



import os
import time
import resource

import numpy as np
import torch

from aenet_gpr.inout.input_parameter import InputParameters
from aenet_gpr.util import ReferenceData, AdditionalData  # , ReferenceDataInternal, AdditionalDataInternal
from aenet_gpr.inout.io_print import *
from aenet_gpr.util.prepare_data import standard_output, inverse_standard_output


class Train(object):
    def __init__(self, input_param: InputParameters):
        self.input_param = input_param
        self.train_data = None

    def read_reference_train_data(self):
        start = time.time()
        if self.input_param.descriptor == "internal":
            pass
            # self.train_data = ReferenceDataInternal(structure_files=self.input_param.train_file,
            #                                         file_format=self.input_param.file_format,
            #                                         device=self.input_param.device,
            #                                         descriptor=self.input_param.descriptor,
            #                                         standardization=self.input_param.standardization,
            #                                         data_type=self.input_param.data_type,
            #                                         data_process=self.input_param.data_process,
            #                                         soap_param=self.input_param.soap_param,
            #                                         mask_constraints=self.input_param.mask_constraints)

        else:
            self.train_data = ReferenceData(structure_files=self.input_param.train_file,
                                            file_format=self.input_param.file_format,
                                            device=self.input_param.device,
                                            descriptor=self.input_param.descriptor,
                                            standardization=self.input_param.standardization,
                                            data_type=self.input_param.data_type,
                                            data_process=self.input_param.data_process,
                                            soap_param=self.input_param.soap_param,
                                            mask_constraints=self.input_param.mask_constraints)

        io_data_read_finalize(t=start,
                              mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                              mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3,
                              flag='training',
                              energy_shape=self.train_data.energy.shape,
                              force_shape=self.train_data.force.shape)

    def train_model(self):
        start = time.time()

        threshold = self.input_param.filter_threshold

        if self.input_param.filter:
            self.train_data.filter_similar_data(threshold=threshold)

        if self.train_data.standardization:
            self.train_data.standardize_energy_force(self.train_data.energy)

        # try:
        if self.input_param.descriptor == "internal":
            pass
            # self.train_data.config_calculator(kerneltype=self.input_param.kerneltype,
            #                                   weight=self.input_param.weight,
            #                                   noise=self.input_param.noise,
            #                                   noisefactor=self.input_param.noisefactor,
            #                                   use_forces=self.input_param.use_forces,
            #                                   sparse=self.input_param.sparse,
            #                                   sparse_derivative=self.input_param.sparse_derivative,
            #                                   autograd=self.input_param.autograd,
            #                                   train_batch_size=self.input_param.train_batch_size,
            #                                   eval_batch_size=self.input_param.eval_batch_size,
            #                                   fit_weight=self.input_param.fit_weight,
            #                                   fit_scale=self.input_param.fit_scale)

        else:
            self.train_data.config_calculator(kerneltype=self.input_param.kerneltype,
                                              scale=self.input_param.scale,
                                              weight=self.input_param.weight,
                                              noise=self.input_param.noise,
                                              noisefactor=self.input_param.noisefactor,
                                              use_forces=self.input_param.use_forces,
                                              sparse=self.input_param.sparse,
                                              sparse_derivative=self.input_param.sparse_derivative,
                                              autograd=self.input_param.autograd,
                                              train_batch_size=self.input_param.train_batch_size,
                                              eval_batch_size=self.input_param.eval_batch_size,
                                              fit_weight=self.input_param.fit_weight,
                                              fit_scale=self.input_param.fit_scale)

        # if self.train_data.calculator.weight < max_weight:
        #     break
        # else:
        #     raise ValueError(f"Weight parameter too high ({self.train_data.calculator.weight}).")
        #
        # except Exception as e:
        #     print(f"{e} Fix the weight and scale parameters.")
        #     self.input_param.fit_weight = False
        #     self.input_param.fit_scale = False

        io_train_finalize(t=start,
                          mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                          mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3,
                          data_param=self.train_data.write_params())

    def train_model_save(self):
        start = time.time()
        self.train_data.save_data()
        io_train_model_save(t=start,
                            mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                            mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3)

    def write_reference_train_xsf(self):
        start = time.time()
        if not os.path.exists("train_xsf"):
            os.makedirs("train_xsf")
        self.train_data.write_image_xsf(path="train_xsf")
        io_train_write_finalize(t=start,
                                mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                                mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3,
                                path="train_xsf")


class Test(object):
    def __init__(self, input_param: InputParameters):
        self.input_param = input_param
        self.train_data = None
        self.test_data = None

    def load_train_model(self, train_data: ReferenceData = None):
        start = time.time()
        if train_data is not None:
            self.train_data = train_data
        else:
            self.train_data = ReferenceData()
            self.train_data.load_data()
            io_train_model_load(t=start,
                                mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                                mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3)

    def read_reference_test_data(self):
        start = time.time()
        if self.input_param.descriptor == "internal":
            pass
            # self.test_data = ReferenceDataInternal(structure_files=self.input_param.test_file,
            #                                        file_format=self.input_param.file_format,
            #                                        device=self.input_param.device,
            #                                        descriptor=self.input_param.descriptor,
            #                                        standardization=self.input_param.standardization,
            #                                        data_type=self.input_param.data_type,
            #                                        data_process=self.input_param.data_process,
            #                                        soap_param=self.input_param.soap_param,
            #                                        mask_constraints=self.input_param.mask_constraints,
            #                                        c_table=self.train_data.c_table)

        else:
            self.test_data = ReferenceData(structure_files=self.input_param.test_file,
                                           file_format=self.input_param.file_format,
                                           device=self.input_param.device,
                                           descriptor=self.input_param.descriptor,
                                           standardization=self.input_param.standardization,
                                           data_type=self.input_param.data_type,
                                           data_process=self.input_param.data_process,
                                           soap_param=self.input_param.soap_param,
                                           mask_constraints=self.input_param.mask_constraints)

        if self.test_data.energy is not None:
            test_data_energy_shape = self.test_data.energy.shape
        else:
            test_data_energy_shape = None

        if self.test_data.force is not None:
            test_data_force_shape = self.test_data.force.shape
        else:
            test_data_force_shape = None
        io_data_read_finalize(t=start,
                              mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                              mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3,
                              flag='test',
                              energy_shape=test_data_energy_shape,
                              force_shape=test_data_force_shape)

    def model_test_evaluation(self):
        start = time.time()
        self.test_data.calculator = self.train_data.calculator
        energy_test_gpr, force_test_gpr, uncertainty_test_gpr = self.test_data.evaluation(
            get_variance=self.input_param.get_variance)

        if self.train_data.standardization:
            energy_test_gpr, force_test_gpr = inverse_standard_output(energy_ref=self.train_data.energy,
                                                                      scaled_energy_target=energy_test_gpr,
                                                                      scaled_force_target=force_test_gpr)
        else:
            pass

        io_test_evaluation(t=start,
                           mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                           mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3,
                           data_param=self.test_data.write_params())

        if self.test_data.energy is not None:
            abs_F_test_gpr = np.linalg.norm(force_test_gpr, axis=2)
            abs_F_test = np.linalg.norm(self.test_data.force, axis=2)

            print("GPR energy MAE (eV):", np.absolute(np.subtract(energy_test_gpr, self.test_data.energy)).mean())
            print("GPR force MAE (eV/Ang):", np.absolute(np.subtract(abs_F_test_gpr, abs_F_test)).mean())
            print(
                "GPR uncertainty mean ± std: {0} ± {1}".format(uncertainty_test_gpr.mean(), uncertainty_test_gpr.std()))

            print("")
            print("Saving test target to [energy_test_reference.npy] and [force_test_reference.npy]")
            np.save("./energy_test_reference.npy", self.test_data.energy)
            np.save("./force_test_reference.npy", self.test_data.force)
        else:
            pass

        if self.input_param.get_variance:
            print(
                "Saving GPR prediction to [energy_test_gpr.npy], [force_test_gpr.npy], and [uncertainty_test_gpr.npy]")
            np.save("./energy_test_gpr.npy", energy_test_gpr)
            np.save("./force_test_gpr.npy", force_test_gpr)
            np.save("./uncertainty_test_gpr.npy", uncertainty_test_gpr)
        else:
            print("Saving GPR prediction to [energy_test_gpr.npy] and [force_test_gpr.npy]")
            np.save("./energy_test_gpr.npy", energy_test_gpr)
            np.save("./force_test_gpr.npy", force_test_gpr)
        print("")
        print("")

    def write_reference_test_xsf(self):
        start = time.time()
        if not os.path.exists("test_xsf"):
            os.makedirs("test_xsf")
        self.test_data.write_image_xsf(path="test_xsf")
        io_test_write_finalize(t=start,
                               mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                               mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3,
                               path="test_data")


class Augmentation(object):
    def __init__(self, input_param: InputParameters):
        self.input_param = input_param
        self.train_data = None
        self.additional_data = None

    def load_train_model(self, train_data: ReferenceData = None):
        start = time.time()
        if train_data is not None:
            self.train_data = train_data
        else:
            self.train_data = ReferenceData()
            self.train_data.load_data()
            io_train_model_load(t=start,
                                mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                                mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3)

    def generate_additional_structures(self):
        start = time.time()
        if self.input_param.descriptor == "internal":
            pass
            # self.additional_data = AdditionalDataInternal(reference_training_data=self.train_data,
            #                                               disp_length=self.input_param.disp_length,
            #                                               num_copy=self.input_param.num_copy)

        else:
            self.additional_data = AdditionalData(reference_training_data=self.train_data,
                                                  disp_length=self.input_param.disp_length,
                                                  num_copy=self.input_param.num_copy)

        self.additional_data.generate_additional_image()
        io_additional_generate_finalize(t=start,
                                        mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                                        mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3,
                                        disp_length=self.additional_data.disp_length,
                                        num_copy=self.additional_data.num_copy,
                                        num_reference=len(self.train_data.images),
                                        num_additional=len(self.additional_data.additional_images))

    def model_additional_evaluation(self):
        start = time.time()
        self.additional_data.evaluation_additional(get_variance=self.input_param.get_variance)
        io_additional_evaluation(t=start,
                                 mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                                 mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3,
                                 data_param=self.additional_data.write_params())

        print("")
        if self.input_param.get_variance:
            print(
                "Saving GPR prediction to [energy_additional_gpr.npy], [force_additional_gpr.npy], and [uncertainty_additional_gpr.npy]")
            np.save("./energy_additional_gpr.npy", self.additional_data.energy_additional)
            np.save("./force_additional_gpr.npy", self.additional_data.force_additional)
            np.save("./uncertainty_additional_gpr.npy", self.additional_data.uncertainty_additional)
        else:
            print("Saving GPR prediction to [energy_additional_gpr.npy] and [force_additional_gpr.npy]")
            np.save("./energy_additional_gpr.npy", self.additional_data.energy_additional)
            np.save("./force_additional_gpr.npy", self.additional_data.force_additional)

    def write_additional_xsf(self):
        start = time.time()
        if not os.path.exists("additional_xsf"):
            os.makedirs("additional_xsf")
        self.additional_data.write_additional_image_xsf(path="additional_xsf")
        io_additional_write_finalize(t=start,
                                     mem_CPU=resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024 ** 2,
                                     mem_GPU=torch.cuda.max_memory_allocated() / 1024 ** 3,
                                     path="additional_xsf")



import os
import copy
import glob
import numpy as np

# import chemcoord as cc
import ase.io
import torch

from aenet_gpr.util.reference_data_internal import ReferenceDataInternal
from aenet_gpr.util.prepare_data import standard_output, inverse_standard_output


class AdditionalDataInternal(object):
    def __init__(self, reference_training_data: ReferenceDataInternal, disp_length=0.055, num_copy=36):

        self.reference_training_data = reference_training_data
        self.data_type = self.reference_training_data.data_type
        self.numpy_data_type = self.reference_training_data.numpy_data_type
        self.torch_data_type = self.reference_training_data.torch_data_type
        self.device = reference_training_data.device

        self.disp_length = disp_length
        self.num_copy = num_copy

        self.species = self.reference_training_data.species
        self.pbc = self.reference_training_data.pbc
        self.num_atom = self.reference_training_data.num_atom
        self.fix_ind = self.reference_training_data.fix_ind
        self.descriptor = self.reference_training_data.descriptor

        self.mask_constraints = self.reference_training_data.mask_constraints
        self.atoms_mask = self.reference_training_data.atoms_mask

        self.additional_images = []
        self.energy_additional = np.array([], dtype=self.numpy_data_type)  # [Ndata]
        self.force_additional = np.array([], dtype=self.numpy_data_type)  # [Ndata, Natom, 3]
        self.uncertainty_additional = np.array([], dtype=self.numpy_data_type)  # [Ndata]

        self.energy_scale_additional = np.array([], dtype=self.numpy_data_type)  # [Ndata]
        self.force_scale_additional = np.array([], dtype=self.numpy_data_type)  # [Ndata, Natom, 3]

    def generate_additional_image(self):

        for i, image in enumerate(self.reference_training_data.images):
            pos_ref = image.get_positions()
            rand_disps = np.random.normal(0, self.disp_length, size=(self.num_copy, self.num_atom, 3))

            for j, rand_disp in enumerate(rand_disps):
                image_disp = copy.deepcopy(image)

                rand_disp[:, 0] = rand_disp[:, 0] - rand_disp[:, 0].mean()
                rand_disp[:, 1] = rand_disp[:, 1] - rand_disp[:, 1].mean()
                rand_disp[:, 2] = rand_disp[:, 2] - rand_disp[:, 2].mean()

                if self.reference_training_data.fix_ind is not None:
                    rand_disp[self.reference_training_data.fix_ind] = np.array([0.0, 0.0, 0.0])

                # set displaced atomic positions
                pos_new = pos_ref + rand_disp
                image_disp.set_positions(pos_new)

                self.additional_images.append(image_disp)

    def write_params(self):
        return dict(num_data=len(self.additional_images),
                    calculator=self.reference_training_data.calculator.hyper_params,
                    fix_ind=self.fix_ind,
                    pbc=self.pbc,
                    species=self.species,
                    num_atom=self.num_atom)

    def evaluation_additional(self, get_variance=False):

        self.reference_training_data.calculator.eval()
        with torch.no_grad():
            if get_variance:
                if self.reference_training_data.standardization:
                    energy_scale_additional, force_scale_additional, uncertainty_additional = self.reference_training_data.calculator(
                        eval_images=self.additional_images,
                        get_variance=get_variance)

                    self.energy_scale_additional = energy_scale_additional.cpu().detach().numpy()
                    self.force_scale_additional = force_scale_additional.cpu().detach().numpy()
                    self.uncertainty_additional = uncertainty_additional.cpu().detach().numpy()

                    self.inverse_standardize_energy_force(reference_training_energy=self.reference_training_data.energy)

                else:
                    energy_additional, force_additional, uncertainty_additional = self.reference_training_data.calculator(
                        eval_images=self.additional_images,
                        get_variance=get_variance)

                    self.energy_additional = energy_additional.cpu().detach().numpy()
                    self.force_additional = force_additional.cpu().detach().numpy()
                    self.uncertainty_additional = uncertainty_additional.cpu().detach().numpy()

            else:
                if self.reference_training_data.standardization:
                    energy_scale_additional, force_scale_additional, _ = self.reference_training_data.calculator(
                        eval_images=self.additional_images,
                        get_variance=get_variance)

                    self.energy_scale_additional = energy_scale_additional.cpu().detach().numpy()
                    self.force_scale_additional = force_scale_additional.cpu().detach().numpy()

                    self.inverse_standardize_energy_force(reference_training_energy=self.reference_training_data.energy)

                else:
                    energy_additional, force_additional, _ = self.reference_training_data.calculator(
                        eval_images=self.additional_images,
                        get_variance=get_variance)

                    self.energy_additional = energy_additional.cpu().detach().numpy()
                    self.force_additional = force_additional.cpu().detach().numpy()

    def standardize_energy_force(self, reference_training_energy):
        """
        Y = [n_systems]
        dY = [n_systems, Natom, 3]
        :return:
        """
        self.energy_scale_additional, self.force_scale_additional = standard_output(reference_training_energy,
                                                                                    self.energy_additional,
                                                                                    self.force_additional)

    def inverse_standardize_energy_force(self, reference_training_energy):
        """
        Y = [n_systems]
        dY = [n_systems, Natom, 3]
        :return:
        """
        self.energy_additional, self.force_additional = inverse_standard_output(reference_training_energy,
                                                                                self.energy_scale_additional,
                                                                                self.force_scale_additional)

    def write_additional_image_xsf(self, path):

        i = 0
        for image in self.additional_images:
            ase.io.write(os.path.join(path, "file_{0:0>5}.xsf".format(i)), image, format="xsf")
            i = i + 1

        files = glob.glob(os.path.join(path, "file_*.xsf"))
        files.sort()

        for i, file in enumerate(files):
            with open(file, 'r') as infile:
                lines = infile.readlines()

            if self.pbc:
                new_lines = lines[:7]
                del lines[:7]
            else:
                new_lines = [lines[0]]
                del lines[0]

            for j, line in enumerate(lines):
                tmp = line.split()
                tmp[0] = self.species[j]

                try:
                    tmp[4] = "%16.14f" % self.force_additional[i, j, 0]
                    tmp[5] = "%16.14f" % self.force_additional[i, j, 1]
                    tmp[6] = "%16.14f" % self.force_additional[i, j, 2]
                except:
                    tmp.append("%16.14f" % self.force_additional[i, j, 0])
                    tmp.append("%16.14f" % self.force_additional[i, j, 1])
                    tmp.append("%16.14f" % self.force_additional[i, j, 2])

                new_line = "     ".join(tmp)
                new_lines.append(new_line + "\n")

            with open(file, "w") as outfile:
                comment = "# total energy = %20.16f eV\n\n" % float(self.energy_additional[i])
                outfile.write("%s" % comment)

                for new_line in new_lines:
                    outfile.write(new_line)




"""
aenet-gpr: A Python package for Gaussian Process Regression (GPR) surrogate modeling
to augment energy data for GPR-ANN potential training
"""
from .inout import inout_process, input_parameter, io_print, read_input
from .src import calculator, gpr_batch, gpr_batch_internal, gpr_iterative, prior, pytorch_kernel, pytorch_kernel_internal, pytorch_kerneltypes
from .util import additional_data, param_optimization, prepare_data, reference_data, reference_data_internal
from .tool import acquisition, aidneb, ase_tool, trainingset


__version__ = "2.2.2"
__all__ = ["inout", "inout_process", "input_parameter", "io_print", "read_input",
           "src", "calculator", "gpr_batch", "gpr_batch_internal", "gpr_iterative", "prior", "pytorch_kernel", "pytorch_kernel_internal", "pytorch_kerneltypes",
           "util", "additional_data", "param_optimization", "prepare_data", "reference_data", "reference_data_internal",
           "tool", "acquisition", "aidneb", "ase_tool", "trainingset"]



import torch
import numpy as np


def get_N_batch(len_dataset, batch_size):
    """
    Returns the number of batches for a given batch size and dataset size
    """
    N_batch = int(len_dataset / batch_size)
    residue = len_dataset - N_batch * batch_size

    if residue >= int(batch_size / 2) or N_batch == 0:
        if residue != 0:
            N_batch += 1

    return N_batch


def get_batch_indexes_N_batch(len_dataset, N_batch):
    """
    Returns the indexes of the structures in StructureDataset that belong to each batch
    """
    finish = 0
    indexes = []

    base, extra = divmod(len_dataset, N_batch)
    N_per_batch = base + (torch.arange(N_batch) < extra).long()

    for i in range(N_batch):
        start = finish
        finish = start + N_per_batch[i]
        indexes.append([start, finish])

    return indexes


def standard_output(energy_ref, energy_target, force_target):
    """
    energy_ref: (Ndata,) numpy array
    energy_target: (Ndata,) numpy array
    force_target: (Ndata, Natom, 3) numpy array

    Returns:
        scaled_energy_target: (Ndata,) numpy array of standardized energy
        scaled_force_target: (Ndata, Natom, 3) numpy array of standardized atomic forces
    """
    n_system = force_target.shape[0]
    n_atom = force_target.shape[1]

    # Mean and std of energy (population standard deviation)
    mean_energy = np.mean(energy_ref)
    std_energy = np.std(energy_ref)

    # Standardize: (energy_target - mean) / std
    scaled_energy_target = (energy_target - mean_energy) / std_energy

    # Average of force is approximately 0
    scaled_force_target = force_target.reshape(n_system, -1) / std_energy
    scaled_force_target = scaled_force_target.reshape(n_system, n_atom, 3)

    return scaled_energy_target, scaled_force_target


def inverse_standard_output(energy_ref, scaled_energy_target, scaled_force_target):
    """
    energy_ref: (Ndata,) numpy array of reference energy
    scaled_energy_target: (Ndata,) numpy array of standardized energy
    scaled_force_target: (Ndata, Natom, 3) numpy array of standardized atomic forces

    Returns:
        energy_target: (Ndata,) numpy array of unscaled energy
        force_target: (Ndata, Natom, 3) numpy array of unscaled atomic forces
    """
    n_system = scaled_force_target.shape[0]
    n_atom = scaled_force_target.shape[1]

    # Mean and std of energy (population standard deviation)
    mean_energy = np.mean(energy_ref)
    std_energy = np.std(energy_ref)

    # Restore Energy: scaled_energy_target * std + mean
    energy_target = scaled_energy_target * std_energy + mean_energy

    # Restore Force: scaled_force_target * std
    force_target = scaled_force_target.reshape(n_system, -1) * std_energy
    force_target = force_target.reshape(n_system, n_atom, 3)

    return energy_target, force_target



import torch
import numpy as np
import gc

from dscribe.descriptors import SOAP

from aenet_gpr.src.pytorch_kerneltypes import SquaredExp
from aenet_gpr.util.prepare_data import get_N_batch, get_batch_indexes_N_batch


class BaseKernelType:
    '''
    Base class for all kernel types with common properties,
    attributes and methods.
    '''

    def __init__(self, params=None):
        # Currently, all the kernel types take weight and scale as parameters

        if params is None:
            self.params = {}

    @property
    def scale(self):
        return self.params['scale']

    @property
    def weight(self):
        return self.params['weight']

    def update(self, params):
        '''
        Update the kernel function hyperparameters.
        '''

        self.params.update(params)


class FPKernel(BaseKernelType):

    def __init__(self,
                 species, pbc, Natom,
                 kerneltype='sqexp',
                 params=None,
                 data_type='float64',
                 soap_param=None,
                 descriptor='cartesian coordinates',
                 device='cpu',
                 atoms_mask=None):
        super().__init__()
        '''
        params: dict
            Hyperparameters for the kernel type
        '''
        kerneltypes = {'sqexp': SquaredExp}
        self.device = device
        self.atoms_mask = atoms_mask

        if data_type == 'float32':
            self.data_type = 'float32'
            self.torch_data_type = torch.float32
        else:
            self.data_type = 'float64'
            self.torch_data_type = torch.float64

        self.species = species
        self.pbc = pbc
        self.Natom = Natom

        if self.atoms_mask is not None:
            self.Nmask = self.atoms_mask.shape[0]
            self.free_atoms = self.atoms_mask // 3
            self.free_atoms = torch.unique(self.free_atoms)
        else:
            self.Nmask = 3 * self.Natom

        if params is None:
            params = {}

        kerneltype = kerneltypes.get(kerneltype)
        self.kerneltype = kerneltype(**params)

        self.soap_param = soap_param
        self.descriptor = descriptor
        self.soap = None

        if self.descriptor == 'soap':
            self.set_soap(r_cut=self.soap_param.get('r_cut'),
                          n_max=self.soap_param.get('n_max'),
                          l_max=self.soap_param.get('l_max'),
                          sigma=self.soap_param.get('sigma'),
                          rbf=self.soap_param.get('rbf'),
                          sparse=self.soap_param.get('sparse'))

    def pairwise_distances(self, fingerprints, sin_transform=False):
        # fingerprints: (Ndata, Ncenter, Nfeature)
        Ndata = fingerprints.shape[0]

        distances = []
        for i in range(Ndata):
            for j in range(i + 1, Ndata):
                diff = fingerprints[i].flatten() - fingerprints[j].flatten()
                if sin_transform:
                    diff = torch.sin(diff / 2)
                distance = torch.linalg.norm(diff)
                distances.append(distance.item())

        distances = torch.tensor(distances)
        mean_distance = torch.mean(distances)
        std_distance = torch.std(distances)
        max_distance = torch.max(distances)
        min_distance = torch.min(distances)

        return mean_distance, std_distance, max_distance, min_distance

    def set_soap(self, r_cut=5.0, n_max=6, l_max=4, sigma=0.5, rbf='gto', sparse=False):
        self.soap = SOAP(
            species=set(self.species),
            periodic=self.pbc,
            r_cut=r_cut,
            n_max=n_max,
            l_max=l_max,
            sigma=sigma,
            rbf=rbf,
            dtype=self.data_type,
            sparse=sparse)

    def generate_descriptor(self, images):

        if self.descriptor == 'soap':
            dfp_dr, fp = self.soap.derivatives(images,
                                               centers=[self.soap_param.get('centers')] * len(images),
                                               method=self.soap_param.get('method'),
                                               return_descriptor=True,
                                               n_jobs=self.soap_param.get('n_jobs'))

            if self.atoms_mask is not None:
                dfp_dr = torch.as_tensor(dfp_dr, dtype=self.torch_data_type).to(self.device)[:, :, :, :, self.atoms_mask]  # (Ndata, Ncenters, Natom, 3, Natoms_masked)
                dfp_dr = dfp_dr[:, :, self.free_atoms, :, :]  # (Ndata, Ncenters, Natom_masked, 3, Natoms_masked)
                fp = torch.as_tensor(fp, dtype=self.torch_data_type).to(self.device)[:, :, self.atoms_mask]  # (Ndata, Ncenters, Natoms_masked)
            else:
                dfp_dr = torch.as_tensor(dfp_dr, dtype=self.torch_data_type).to(self.device)  # (Ndata, Ncenters, Natom, 3, Natom*3)
                fp = torch.as_tensor(fp, dtype=self.torch_data_type).to(self.device)  # (Ndata, Ncenters, Natom*3)

        else:
            fp = []
            dfp_dr = []
            for image in images:
                fp.append(torch.as_tensor(image.get_positions(wrap=False).reshape(-1), dtype=self.torch_data_type).to(self.device))

                dfp_dr.append(torch.as_tensor(np.eye(self.Natom * 3).reshape(self.Natom, -1, 3, order='F'),
                                              dtype=self.torch_data_type).to(self.device))

            fp = torch.stack(fp).to(self.device)  # (Ndata, Natom*3)
            dfp_dr = torch.stack(dfp_dr).to(self.device)  # (Ndata, Natom, Natom*3, 3)

            if self.atoms_mask is not None:
                fp = fp.unsqueeze(1)[:, :, self.atoms_mask]  # (Ndata, 1, Natoms_masked)
                dfp_dr = dfp_dr.transpose(2, 3).unsqueeze(1)[:, :, :, :, self.atoms_mask]  # (Ndata, 1, Natom, 3, Natoms_masked)
                dfp_dr = dfp_dr[:, :, self.free_atoms, :, :]  # (Ndata, Ncenters, Natom_masked, 3, Natoms_masked)
            else:
                fp = fp.unsqueeze(1)  # (Ndata, 1, Natom*3)
                dfp_dr = dfp_dr.transpose(2, 3).unsqueeze(1)  # (Ndata, 1, Natom, 3, Natom*3)

        return fp, dfp_dr

    def generate_descriptor_per_data(self, image):

        if self.descriptor == 'soap':
            dfp_dr, fp = self.soap.derivatives(image,
                                               centers=self.soap_param.get('centers'),
                                               method=self.soap_param.get('method'),
                                               return_descriptor=True,
                                               n_jobs=self.soap_param.get('n_jobs'))

            if self.atoms_mask is not None:
                dfp_dr = torch.as_tensor(dfp_dr, dtype=self.torch_data_type).to(self.device)[:, :, :, self.atoms_mask]  # (Ncenters, Natom, 3, Natoms_masked)
                dfp_dr = dfp_dr[:, self.free_atoms, :, :]  # (Ncenters, Natom_masked, 3, Natoms_masked)
                fp = torch.as_tensor(fp, dtype=self.torch_data_type).to(self.device)[:, self.atoms_mask]  # (Ncenters, Natoms_masked)
            else:
                dfp_dr = torch.as_tensor(dfp_dr, dtype=self.torch_data_type).to(self.device)  # (Ncenters, Natom, 3, Natom*3)
                fp = torch.as_tensor(fp, dtype=self.torch_data_type).to(self.device)  # (Ncenters, Natom*3)

        else:
            fp = torch.as_tensor(image.get_positions(wrap=False).reshape(-1), dtype=self.torch_data_type).to(self.device)
            dfp_dr = torch.as_tensor(np.eye(self.Natom * 3).reshape(self.Natom, -1, 3, order='F'),
                                     dtype=self.torch_data_type).to(self.device)

            if self.atoms_mask is not None:
                fp = fp.unsqueeze(0)[:, self.atoms_mask]
                dfp_dr = dfp_dr.transpose(1, 2).unsqueeze(0)[:, :, :, self.atoms_mask]
                dfp_dr = dfp_dr[:, self.free_atoms, :, :]
            else:
                fp = fp.unsqueeze(0)
                dfp_dr = dfp_dr.transpose(1, 2).unsqueeze(0)

        return fp, dfp_dr

    def kernel_with_deriv(self, X1=None, dX1=None, X2=None, dX2=None):
        '''
        Data process type: Broadcast, Batch Processing
        Return a full kernel matrix between two structural fingerprints, 'X1' and 'X2'.
        X1, X2 = [Ncenter, Nfeature] -> [Ndata, Ncenter, Nfeature]
        dX1, dX2 = [Ncenter, Natom, 3, Nfeature] -> [Ndata, Ncenter, Natom, 3, Nfeature]
        where Ndata = Ntrain or Nsparse or Ntest
        '''

        # Create empty kernel
        # [[K(X11,X21), K(X11,X22), ..., K(X11,X2N), K(X11,dX21x), K(X11,dX21y), ..., K(X11,dX2Nz)],
        #  [K(X12,X21), K(X12,X22), ..., K(X12,X2N), K(X12,dX21x), K(X12,dX21y), ..., K(X12,dX2Nz)]
        #  ...
        #  [K(X1N,X21), K(X1N,X22), ..., K(X1N,X2N), K(X1N,dX21x), K(X1N,dX21y), ..., K(X1N,dX2Nz)]
        #  [K(dX11x,X21), K(dX11x,X22), ..., K(dX11x,X2N), K(dX11x,dX21x), K(dX11x,dX21y), ..., K(dX11x,dX2Nz)]
        #  ...
        #  [K(dX1Nz,X21), K(dX1Nz,X22), ..., K(dX1Nz,X2N), K(dX1Nz,dX21x), K(dX1Nz,dX21y), ..., K(dX1Nz,dX2Nz)]

        K_X1X2 = torch.empty(((1 + self.Nmask) * X1.shape[0],
                              (1 + self.Nmask) * X2.shape[0]), dtype=self.torch_data_type, device=self.device)

        # Expand value fingerprint
        X1_expanded = X1[:, None, :, None, :]  # [Ndata1, 1, Ncenter, 1, Nfeature]
        X2_expanded = X2[None, :, None, :, :]  # [1, Ndata2, 1, Ncenter, Nfeature]

        X1__outer_minus__X2 = X1_expanded - X2_expanded  # [Ndata1, Ndata2, Ncenter, Ncenter, Nfeature]
        X2__outer_minus__X1 = X2_expanded - X1_expanded  # [Ndata1, Ndata2, Ncenter, Ncenter, Nfeature]

        # [Ndata, Ndata, Ncenter, Ncenter]
        __k_X1X2 = self.weight ** 2 * torch.exp(
            -torch.linalg.norm(X1__outer_minus__X2, dim=4) ** 2 / (2 * self.scale ** 2))
        __k_X2X1 = self.weight ** 2 * torch.exp(
            -torch.linalg.norm(X2__outer_minus__X1, dim=4) ** 2 / (2 * self.scale ** 2))

        """"K_X1X2[:X1.shape[0], :X2.shape[0]] kernel between fp1 and fp2"""
        # kernel = torch.sum(__k_X1X2, dim=(2, 3))
        # K_X1X2[:X1.shape[0], :X2.shape[0]].copy_(kernel)
        K_X1X2[:X1.shape[0], :X2.shape[0]] = torch.sum(__k_X1X2, dim=(2, 3))

        """K_X1X2[X1.shape[0]: X1.shape[0] * (1 + 3 * Natom), :X2.shape[0]] kernel between fp1_deriv and fp2"""
        # [Ndata, Ncenter, Natom, 3, Nfeature] -> [Ndata, Ncenter, Natom * 3, Nfeature]
        dX1_reshaped = dX1.flatten(start_dim=2, end_dim=3)
        # intermediate_result = torch.einsum('xycnf,xcdf->xycnd', X1__outer_minus__X2, dX1_reshaped)
        # intermediate_result = torch.matmul(X1__outer_minus__X2, dX1_reshaped.transpose(-1, -2))

        # [Ndata1, Ndata2, Natom * 3]
        # kernel = torch.einsum('xycn,xycnd->xyd', __k_X1X2 / self.scale**2, intermediate_result)
        # kernel = torch.matmul(__k_X1X2 / self.scale ** 2, intermediate_result).sum(dim=(2, 3))

        # [Ndata1 * Natom * 3, Ndata2]
        # K_X1X2[X1.shape[0]:, :X2.shape[0]] = kernel.permute(0, 2, 1).reshape(X1.shape[0] * 3 * self.Natom, X2.shape[0])
        # K_X1X2[X1.shape[0]:, :X2.shape[0]].copy_(
        #     kernel.permute(0, 2, 1).contiguous().view(X1.shape[0] * 3 * self.Natom, X2.shape[0]))
        K_X1X2[X1.shape[0]:, :X2.shape[0]].copy_(torch.einsum('xycn,xycnd->xyd', __k_X1X2 / self.scale ** 2,
                                                              torch.einsum('xycnf,xcdf->xycnd', X1__outer_minus__X2,
                                                                           dX1_reshaped)).permute(0, 2,
                                                                                                  1).contiguous().view(
            X1.shape[0] * self.Nmask, X2.shape[0]))

        """K_X1X2[:X1.shape[0], X2.shape[0]: X2.shape[0] * (1 + 3 * Natom)] kernel between fp1 and fp2_deriv"""
        # [Ndata, Ncenter, Natom, 3, Nfeature] -> [Ndata, Ncenter, Natom * 3, Nfeature]
        dX2_reshaped = dX2.flatten(start_dim=2, end_dim=3)
        # intermediate_result = torch.einsum('xycnf,yndf->xycnd', X2__outer_minus__X1, dX2_reshaped)
        # intermediate_result = torch.matmul(X2__outer_minus__X1, dX2_reshaped.transpose(-1, -2))

        # kernel = torch.einsum('xycn,xycnd->xyd', __k_X2X1 / self.scale**2, intermediate_result)
        # kernel = torch.matmul(__k_X2X1 / self.scale ** 2, intermediate_result).sum(dim=(2, 3))

        # [Ndata1, Ndata2 * Natom * 3]
        # K_X1X2[:X1.shape[0], X2.shape[0]:] = kernel.reshape(X1.shape[0], X2.shape[0] * 3 * self.Natom)
        # K_X1X2[:X1.shape[0], X2.shape[0]:].copy_(
        #     kernel.view(X1.shape[0], X2.shape[0] * 3 * self.Natom))
        K_X1X2[:X1.shape[0], X2.shape[0]:].copy_(torch.einsum('xycn,xycnd->xyd', __k_X2X1 / self.scale ** 2,
                                                              torch.einsum('xycnf,yndf->xycnd', X2__outer_minus__X1,
                                                                           dX2_reshaped)).view(X1.shape[0], X2.shape[
            0] * self.Nmask))

        """K_X1X2[X1.shape[0]: X1.shape[0] * (1 + 3 * Natom), X2.shape[0]: X2.shape[0] * (1 + 3 * Natom)] 
        kernel between fp1_deriv and fp2_deriv"""
        # size sign in eisum:
        # x, y: Ndata
        # c, n: Ncenter
        # b, d: Natom *3
        # f: Nfeature
        # [Ndata, Ndata, Ncenter, Ncenter, Nfeature] * [Ndata, Ndata, Ncenter, Natom * 3, Nfeature]
        # -> [Ndata, Ndata, Ncenter, Ncenter, Natom * 3]
        DdD_dr1 = torch.einsum('xycnf,x...cbf->xycnb',
                               X1__outer_minus__X2,
                               dX1_reshaped)
        # -> [Ndata, Ndata, Ncenter, Ncenter, Natom * 3]
        # DdD_dr1 = torch.matmul(X1__outer_minus__X2,
        #                        dX1_reshaped.transpose(-1, -2)).sum(dim=-1)

        DdD_dr2 = torch.einsum('xycnf,...yndf->xycnd',
                               X2__outer_minus__X1,
                               dX2_reshaped)
        # -> [Ndata, Ndata, Ncenter, Ncenter, Natom * 3]
        # DdD_dr2 = torch.matmul(X2__outer_minus__X1,
        #                        dX2_reshaped.transpose(-1, -2)).sum(dim=-1)

        # [Ndata, Ndata, Natom * 3, Natom * 3]
        # -> [Ndata, Ndata, Ncenter, Ncenter, Natom * 3, Natom * 3]
        # intermediate_result = torch.einsum('xycnb,xycnd->xycnbd', DdD_dr1, DdD_dr2)
        # intermediate_result = torch.matmul(DdD_dr1.unsqueeze(-1), DdD_dr2.unsqueeze(-2))

        # C0 = torch.einsum('xycn,xycnbd->xybd',
        #                   __k_X1X2 / self.scale**4, intermediate_result)
        # -> [Ndata, Ndata, Natom * 3, Natom * 3]
        # C0 = ((__k_X1X2 / self.scale ** 4).unsqueeze(-1).unsqueeze(-1) * intermediate_result).sum(dim=(2, 3))
        # kernel = ((__k_X1X2 / self.scale ** 4).unsqueeze(-1).unsqueeze(-1)
        #           * torch.matmul(DdD_dr1.unsqueeze(-1), DdD_dr2.unsqueeze(-2))).sum(dim=(2, 3))
        # kernel = torch.einsum('xycn,xycnbd->xybd', __k_X1X2 / self.scale**4,
        #                       torch.einsum('xycnb,xycnd->xycnbd', DdD_dr1, DdD_dr2))

        # [Ndata, Ncenter, Natom * 3, Nfeature] * [Ndata, Ncenter, Natom * 3, Nfeature]
        # -> [Ndata, Ndata, Ncenter, Ncenter, Natom * 3, Natom * 3]
        # intermediate_result = torch.einsum('xcbf,yndf->xycnbd', dX1_reshaped, dX2_reshaped)
        # intermediate_result = torch.matmul(dX1_reshaped.unsqueeze(-1), dX2_reshaped.unsqueeze(-2)).sum(dim=-1)

        # C1 = torch.einsum('xycn,xycnbd->xybd',
        #                   __k_X1X2 / self.scale**2, intermediate_result)
        # C1 = ((__k_X1X2 / self.scale ** 2).unsqueeze(-1).unsqueeze(-1) * intermediate_result).sum(dim=(2, 3))

        # kernel = C0 + C1
        # kernel.add_(((__k_X1X2 / self.scale ** 2).unsqueeze(-1).unsqueeze(-1)
        #              * torch.matmul(dX1_reshaped.unsqueeze(-1), dX2_reshaped.unsqueeze(-2)).sum(dim=-1)).sum(dim=(2, 3)))
        # kernel.add_(torch.einsum('xycn,xycnbd->xybd', __k_X1X2 / self.scale**2,
        #                          torch.einsum('xcbf,yndf->xycnbd', dX1_reshaped, dX2_reshaped)))

        # [Ndata1 * Natom * 3, Ndata2 * Natom * 3]
        # K_X1X2[X1.shape[0]:, X2.shape[0]:] = kernel.permute(0, 2, 1, 3).reshape(X1.shape[0] * 3 * self.Natom,
        #                                                                         X2.shape[0] * 3 * self.Natom)
        # K_X1X2[X1.shape[0]:, X2.shape[0]:].copy_(
        #     kernel.permute(0, 2, 1, 3).contiguous().view(X1.shape[0] * 3 * self.Natom,
        #                                                  X2.shape[0] * 3 * self.Natom))

        # intermediate_result = torch.einsum('xycnb,xycnd->xycnbd', DdD_dr1, DdD_dr2)
        K_X1X2[X1.shape[0]:, X2.shape[0]:].copy_(torch.einsum('xycn,xycnbd->xybd', __k_X1X2 / self.scale ** 4,
                                                              torch.einsum('xycnb,xycnd->xycnbd', DdD_dr1,
                                                                           DdD_dr2)).permute(0, 2, 1,
                                                                                             3).contiguous().view(
            X1.shape[0] * self.Nmask, X2.shape[0] * self.Nmask))

        # intermediate_result = torch.einsum('xcbf,yndf->xycnbd', dX1_reshaped, dX2_reshaped)
        K_X1X2[X1.shape[0]:, X2.shape[0]:].add_(torch.einsum('xycn,xycnbd->xybd', __k_X1X2 / self.scale ** 2,
                                                             torch.einsum('xcbf,yndf->xycnbd', dX1_reshaped,
                                                                          dX2_reshaped)).permute(0, 2, 1,
                                                                                                 3).contiguous().view(
            X1.shape[0] * self.Nmask, X2.shape[0] * self.Nmask))

        del X1_expanded, X2_expanded, __k_X1X2, __k_X2X1, X1__outer_minus__X2, X2__outer_minus__X1, \
            DdD_dr1, DdD_dr2, dX1_reshaped, dX2_reshaped
        gc.collect()

        return K_X1X2

    def kernel_per_data(self, X1=None, dX1=None, X2=None, dX2=None, iter=0):
        '''
        Return a full kernel matrix between two structural fingerprints, 'X1' and 'X2'.
        X1, X2 = [Ncenter, Nfeature]
        dX1, dX2 = [Ncenter, Natom, 3, Nfeature]
        '''

        Nfeature = X1.shape[1]
        X1_expanded = X1[:, None, :]  # [Ncenter, 1, Nfeature]
        X2_expanded = X2[None, :, :]  # [1, Ncenter, Nfeature]

        X1_outer_minus_X2 = X1_expanded - X2_expanded  # [Ncenter, Ncenter, Nfeature]
        X2_outer_minus_X1 = X2_expanded - X1_expanded  # [Ncenter, Ncenter, Nfeature]

        # RBF kernel
        k = self.weight ** 2 * torch.exp(-torch.linalg.norm(X1_outer_minus_X2, dim=2) ** 2 / (2 * self.scale ** 2))

        # Create empty kernel
        kernel = torch.empty((1 + self.Nmask,
                              1 + self.Nmask), dtype=self.torch_data_type, device=self.device)

        # Evaluate the first element of kernel
        kernel[0, 0] = torch.sum(k)

        # Evaluate the first column of kernel using einsum
        intermediate_result = torch.einsum('cnf,c...af->cna',
                                           X1_outer_minus_X2,
                                           dX1.reshape(X1.shape[0],
                                                       self.Nmask,
                                                       Nfeature))
        kernel[1:, 0] = torch.einsum('cn,cna->a',
                                     k / self.scale ** 2,
                                     intermediate_result)

        # Evaluate the first row of kernel using einsum
        intermediate_result = torch.einsum('cnf,...naf->cna',
                                           X2_outer_minus_X1,
                                           dX2.reshape(X2.shape[0],
                                                       self.Nmask,
                                                       Nfeature))
        kernel[0, 1:] = torch.einsum('cn,cna->a',
                                     k / self.scale ** 2,
                                     intermediate_result)

        # [Ncenter, Nfeature] * [Ncenter, Natom * 3, Nfeature]
        # -> [Ncenter, Natom * 3]
        DdD_dr1 = torch.einsum('cnf,c...bf->cnb',
                               X1_outer_minus_X2,
                               dX1.reshape(X1.shape[0],
                                           self.Nmask,
                                           Nfeature))
        DdD_dr2 = torch.einsum('cnf,...ndf->cnd',
                               X2_outer_minus_X1,
                               dX2.reshape(X2.shape[0],
                                           self.Nmask,
                                           Nfeature))

        # [Ncenter, Natom * 3] * [Ncenter, Natom * 3]
        # -> [Ncenter, Natom * 3, Natom * 3]
        C0 = torch.einsum('cn,cnbd->bd',
                          k / self.scale ** 4,
                          torch.einsum('cnb,cnd->cnbd',
                                       DdD_dr1,
                                       DdD_dr2))

        # [Ncenter, Natom * 3, Nfeature] * [Ncenter, Natom * 3, Nfeature]
        # -> [Ncenter, Natom * 3, Natom * 3]
        intermediate_result = torch.einsum('cbf,ndf->cnbd',
                                           dX1.reshape(X1.shape[0],
                                                       self.Nmask,
                                                       Nfeature),
                                           dX2.reshape(X2.shape[0],
                                                       self.Nmask,
                                                       Nfeature))
        C1 = torch.einsum('cn,cnbd->bd',
                          k / self.scale ** 2,
                          intermediate_result)
        kernel[1:, 1:] = C0 + C1

        if iter % 10 == 9:
            del X1_expanded, X2_expanded, X1_outer_minus_X2, X2_outer_minus_X1, k, intermediate_result, \
                DdD_dr1, DdD_dr2, C0, C1
            gc.collect()

        return kernel

    def kernel_matrix_batch(self, images, batch_size=25):
        '''
        Data process type: Broadcast, Batch Processing
        Calculates C(X,X) i.e. full kernel matrix for training data.
        X = [Ntrain or Nsparse or Ntest, Ncenter, Nfeature]
        dX = [Ntrain or Nsparse or Ntest, Ncenter, Natom, 3, Nfeature]
        '''

        Ndata = len(images)
        X_N_batch = get_N_batch(Ndata, batch_size)
        X_indexes = get_batch_indexes_N_batch(Ndata, X_N_batch)

        K_XX = torch.empty((Ndata * (1 + self.Nmask),
                            Ndata * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)

        for i in range(0, X_N_batch):
            fp_i, dfp_dr_i = self.generate_descriptor(images=images[X_indexes[i][0]:X_indexes[i][1]])

            row_indexes = torch.arange(X_indexes[i][0], X_indexes[i][1])
            row_deriv_indexes = torch.arange(Ndata + X_indexes[i][0] * self.Nmask,
                                             Ndata + X_indexes[i][1] * self.Nmask)
            selected_rows = torch.cat([row_indexes, row_deriv_indexes])

            K_XX[selected_rows[:, None], selected_rows] = self.kernel_with_deriv(X1=fp_i, dX1=dfp_dr_i,
                                                                                 X2=fp_i, dX2=dfp_dr_i)

            for j in range(i + 1, X_N_batch):
                fp_j, dfp_dr_j = self.generate_descriptor(images=images[X_indexes[j][0]:X_indexes[j][1]])

                col_indexes = torch.arange(X_indexes[j][0], X_indexes[j][1])
                col_deriv_indexes = torch.arange(Ndata + X_indexes[j][0] * self.Nmask,
                                                 Ndata + X_indexes[j][1] * self.Nmask)
                selected_cols = torch.cat([col_indexes, col_deriv_indexes])

                K_XX[selected_rows[:, None], selected_cols] = self.kernel_with_deriv(X1=fp_i, dX1=dfp_dr_i,
                                                                                     X2=fp_j, dX2=dfp_dr_j)
                K_XX[selected_cols[:, None], selected_rows] = K_XX[selected_rows[:, None], selected_cols].T

        return K_XX

    def kernel_matrix_iterative(self, images):

        Ndata = len(images)

        K_XX = torch.empty((Ndata * (1 + self.Nmask),
                            Ndata * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)

        for i in range(0, Ndata):
            fp_i, dfp_dr_i = self.generate_descriptor_per_data(image=images[i])

            row_index = torch.tensor([i])
            row_deriv_index = torch.arange(Ndata + i * self.Nmask, Ndata + (i + 1) * self.Nmask)
            selected_rows = torch.cat([row_index, row_deriv_index])

            K_XX[selected_rows[:, None], selected_rows] = self.kernel_per_data(X1=fp_i, dX1=dfp_dr_i,
                                                                               X2=fp_i, dX2=dfp_dr_i,
                                                                               iter=i)

            for j in range(i + 1, Ndata):
                fp_j, dfp_dr_j = self.generate_descriptor_per_data(image=images[j])

                col_index = torch.tensor([j])
                col_deriv_index = torch.arange(Ndata + j * self.Nmask, Ndata + (j + 1) * self.Nmask)
                selected_cols = torch.cat([col_index, col_deriv_index])

                K_XX[selected_rows[:, None], selected_cols] = self.kernel_per_data(X1=fp_i, dX1=dfp_dr_i,
                                                                                   X2=fp_j, dX2=dfp_dr_j,
                                                                                   iter=j)
                K_XX[selected_cols[:, None], selected_rows] = K_XX[selected_rows[:, None], selected_cols].T

        return K_XX

    def kernel_matrix_per_data(self, image):
        fp_i, dfp_dr_i = self.generate_descriptor_per_data(image=image)

        K_X_i_X_i = self.kernel_per_data(X1=fp_i, dX1=dfp_dr_i,
                                         X2=fp_i, dX2=dfp_dr_i)

        return K_X_i_X_i

    def kernel_vector_batch(self, eval_images, train_images, batch_size=25):
        '''
        Data process type: Broadcast, Batch Processing
        Calculates C(x,X) i.e. the kernel matrix between fingerprint "x" and the training data fingerprints in "X".
        x = [Ntest or Nsparse, Ncenter, Nfeature]
        dx = [Ntest or Nsparse, Ncenter, Natom, 3, Nfeature]
        X = [Ntrain, Ncenter, Nfeature]
        dX = [Ntrain, Ncenter, Natom, 3, Nfeature]
        '''

        Ntest = len(eval_images)
        Ntrain = len(train_images)

        x_N_batch = get_N_batch(Ntest, batch_size)
        x_indexes = get_batch_indexes_N_batch(Ntest, x_N_batch)

        X_N_batch = get_N_batch(Ntrain, batch_size)
        X_indexes = get_batch_indexes_N_batch(Ntrain, X_N_batch)

        K_xX = torch.empty((Ntest * (1 + self.Nmask),
                            Ntrain * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)

        for i in range(0, x_N_batch):
            fp_i, dfp_dr_i = self.generate_descriptor(images=eval_images[x_indexes[i][0]:x_indexes[i][1]])

            row_indexes = torch.arange(x_indexes[i][0], x_indexes[i][1])
            row_deriv_indexes = torch.arange(Ntest + x_indexes[i][0] * self.Nmask,
                                             Ntest + x_indexes[i][1] * self.Nmask)
            selected_rows = torch.cat([row_indexes, row_deriv_indexes])

            for j in range(0, X_N_batch):
                fp_j, dfp_dr_j = self.generate_descriptor(images=train_images[X_indexes[j][0]:X_indexes[j][1]])

                col_indexes = torch.arange(X_indexes[j][0], X_indexes[j][1])
                col_deriv_indexes = torch.arange(Ntrain + X_indexes[j][0] * self.Nmask,
                                                 Ntrain + X_indexes[j][1] * self.Nmask)
                selected_cols = torch.cat([col_indexes, col_deriv_indexes])

                K_xX[selected_rows[:, None], selected_cols] = self.kernel_with_deriv(X1=fp_i, dX1=dfp_dr_i,
                                                                                     X2=fp_j, dX2=dfp_dr_j)

        return K_xX

    def kernel_vector_iterative(self, eval_images, train_images):
        '''
        Data process type: Sequential, iterative Processing
        Calculates C(x,X) i.e. the kernel matrix between fingerprint "x" and the training data fingerprints in "X".
        x = [Ntest or Nsparse, Ncenter, Nfeature]
        dx = [Ntest or Nsparse, Ncenter, Natom, 3, Nfeature]
        X = [Ntrain, Ncenter, Nfeature]
        dX = [Ntrain, Ncenter, Natom, 3, Nfeature]
        '''

        Ntest = len(eval_images)
        Ntrain = len(train_images)

        # if dx is not None and dX is not None:
        K_xX = torch.empty((Ntest * (1 + self.Nmask),
                            Ntrain * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)

        for i in range(0, Ntest):
            fp_i, dfp_dr_i = self.generate_descriptor_per_data(image=eval_images[i])

            row_index = torch.tensor([i])
            row_deriv_index = torch.arange(Ntest + i * self.Nmask, Ntest + (i + 1) * self.Nmask)
            selected_rows = torch.cat([row_index, row_deriv_index])

            for j in range(0, Ntrain):
                fp_j, dfp_dr_j = self.generate_descriptor_per_data(image=train_images[j])

                col_index = torch.tensor([j])
                col_deriv_index = torch.arange(Ntrain + j * self.Nmask, Ntrain + (j + 1) * self.Nmask)
                selected_cols = torch.cat([col_index, col_deriv_index])

                K_xX[selected_rows[:, None], selected_cols] = self.kernel_per_data(X1=fp_i, dX1=dfp_dr_i,
                                                                                   X2=fp_j, dX2=dfp_dr_j,
                                                                                   iter=j)

        return K_xX

    def kernel_vector_per_data(self, eval_image, train_images, batch_size=25):
        '''
        Calculates C(x,X) i.e. the kernel matrix between fingerprint "x" and the training data fingerprints in "X".
        x = [Ncenter, Nfeature]
        dx = [Ncenter, Natom, 3, Nfeature]
        X = [Ntrain, Ncenter, Nfeature]
        dX = [Ntrain, Ncenter, Natom, 3, Nfeature]
        '''

        Ntrain = len(train_images)

        X_N_batch = get_N_batch(Ntrain, batch_size)
        X_indexes = get_batch_indexes_N_batch(Ntrain, X_N_batch)

        fp_i, dfp_dr_i = self.generate_descriptor_per_data(image=eval_image)
        fp_i = fp_i.unsqueeze(0)
        dfp_dr_i = dfp_dr_i.unsqueeze(0)

        # K_xX_i = torch.empty(((1 + self.Nmask),
        # Ntrain * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)
        K_xX_i = torch.empty((1 * (1 + self.Nmask),
                              Ntrain * (1 + self.Nmask)), dtype=self.torch_data_type, device=self.device)

        for j in range(0, X_N_batch):
            fp_j, dfp_dr_j = self.generate_descriptor(images=train_images[X_indexes[j][0]:X_indexes[j][1]])

            col_indexes = torch.arange(X_indexes[j][0], X_indexes[j][1])
            col_deriv_indexes = torch.arange(Ntrain + X_indexes[j][0] * self.Nmask,
                                             Ntrain + X_indexes[j][1] * self.Nmask)
            selected_cols = torch.cat([col_indexes, col_deriv_indexes])

            K_xX_i[:, selected_cols] = self.kernel_with_deriv(X1=fp_i, dX1=dfp_dr_i, X2=fp_j, dX2=dfp_dr_j)

        # for j in range(0, Ntrain):
        #     fp_j, dfp_dr_j = self.generate_descriptor_per_data(image=train_images[j])
        #
        #     col_index = torch.tensor([j])
        #     col_deriv_index = torch.arange(Ntrain + j * self.Nmask, Ntrain + (j + 1) * self.Nmask)
        #     selected_cols = torch.cat([col_index, col_deriv_index])
        #
        #     K_xX_i[:, selected_cols] = self.kernel_per_data(X1=fp_i, dX1=dfp_dr_i,
        #                                                     X2=fp_j, dX2=dfp_dr_j, iter=j)

        return K_xX_i

    def set_params(self, params):
        '''
        Set new (hyper)parameters for the kernel function.
        '''
        self.update(params)
        self.kerneltype.update(params)


class FPKernelNoforces(BaseKernelType):

    def __init__(self,
                 species, pbc, Natom,
                 kerneltype='sqexp',
                 params=None,
                 data_type='float64',
                 soap_param=None,
                 descriptor='cartesian coordinates',
                 device='cpu'):
        super().__init__()
        '''
        params: dict
            Hyperparameters for the kernel type
        '''
        kerneltypes = {'sqexp': SquaredExp}
        self.device = device

        if data_type == 'float32':
            self.data_type = 'float32'
            self.torch_data_type = torch.float32
        else:
            self.data_type = 'float64'
            self.torch_data_type = torch.float64

        self.species = species
        self.pbc = pbc
        self.Natom = Natom

        if params is None:
            params = {}

        kerneltype = kerneltypes.get(kerneltype)
        self.kerneltype = kerneltype(**params)

        self.soap_param = soap_param
        self.descriptor = descriptor
        self.soap = None

        if self.descriptor == 'soap':
            self.set_soap(r_cut=self.soap_param.get('r_cut'),
                          n_max=self.soap_param.get('n_max'),
                          l_max=self.soap_param.get('l_max'),
                          sigma=self.soap_param.get('sigma'),
                          rbf=self.soap_param.get('rbf'),
                          sparse=self.soap_param.get('sparse'))

    def set_soap(self, r_cut=5.0, n_max=6, l_max=4, sigma=0.5, rbf='gto', sparse=False):
        self.soap = SOAP(
            species=set(self.species),
            periodic=self.pbc,
            r_cut=r_cut,
            n_max=n_max,
            l_max=l_max,
            sigma=sigma,
            rbf=rbf,
            dtype=self.data_type,
            sparse=sparse)

    def kernel_without_deriv(self, X1=None, X2=None):
        '''
        Data process type: Broadcast, Batch Processing
        Return a full kernel matrix between two structural fingerprints, 'X1' and 'X2'.
        X1, X2 = [Ndata, Ncenter, Nfeature]
        where Ndata = Ntrain or Nsparse
        '''

        X1_expanded = X1[:, None, :, None, :]  # [Ntrain, 1, Ncenter, 1, Nfeature]
        X2_expanded = X2[None, :, None, :, :]  # [1, Ntrain, 1, Ncenter, Nfeature]
        X1__outer_minus__X2 = X1_expanded - X2_expanded

        k = self.weight ** 2 * torch.exp(-torch.linalg.norm(X1__outer_minus__X2, dim=4) ** 2 / (2 * self.scale ** 2))

        kernel = torch.sum(k, dim=(2, 3))

        return kernel

    def kernel_matrix(self, X=None):
        '''
        Data process type: Broadcast, Batch Processing
        Calculates C(X,X) i.e. full kernel matrix for training data.
        X = [Ntrain or Nsparse or Ntest, Ncenter, Nfeature]
        '''

        K_XX = self.kernel_without_deriv(X, X)

        return K_XX

    def kernel_vector(self, x=None, X=None):
        '''
        Data process type: Broadcast, Batch Processing
        Calculates C(x,X) i.e. the kernel matrix between fingerprint "x" and the training data fingerprints in "X".
        x = [Ntest or Nsparse, Ncenter, Nfeature]
        X = [Ntrain, Ncenter, Nfeature]
        '''

        K_xX = self.kernel_without_deriv(x, X)

        return K_xX

    def set_params(self, params):
        '''
        Set new (hyper)parameters for the kernel function.
        '''
        self.update(params)
        self.kerneltype.update(params)



import glob
import ast
from aenet_gpr.inout.input_parameter import InputParameters


def read_keyword_argument_same_line(keyword, lines):
	value = None
	found = False
	for line in lines:
		read_keyword = line.split()[0].lower()

		if read_keyword == keyword:
			found = True
			value = line.split()[1]
	return value, found


def read_keyword_list_same_line(keyword, lines):
	value = None
	found = False
	for line in lines:
		read_keyword = line.split()[0].lower()

		if read_keyword == keyword:
			found = True
			idx = line.find('[')
			if idx != -1:
				list_str = line[idx:]
				value = ast.literal_eval(list_str)
			break
	return value, found


def read_train_in(infile):
	with open(infile, "r") as f:

		# Initialize InputParameters with default values
		input_param = InputParameters()

		# Remove comments from input file:
		lines = f.readlines()
		list_comments = []
		for i in range(len(lines)-1, -1, -1):
			if lines[i][0] in ["!", "#"] or len(lines[i].split()) == 0:
				list_comments.append(i)

		for i in list_comments:
			lines.pop(i)

		# Logical parameters:
		standardization, found = read_keyword_argument_same_line("standardization", lines)
		if found:
			if "F" in standardization.upper():
				input_param.standardization = False
			else:
				input_param.standardization = True

		constraints, found = read_keyword_argument_same_line("constraints", lines)
		if found:
			if "F" in constraints.upper():
				input_param.mask_constraints = False
			else:
				input_param.mask_constraints = True

		fit_weight, found = read_keyword_argument_same_line("fit_weight", lines)
		if found:
			if "F" in fit_weight.upper():
				input_param.fit_weight = False
			else:
				input_param.fit_weight = True

		fit_scale, found = read_keyword_argument_same_line("fit_scale", lines)
		if found:
			if "F" in fit_scale.upper():
				input_param.fit_scale = False
			else:
				input_param.fit_scale = True

		use_forces, found = read_keyword_argument_same_line("force", lines)
		if found:
			if "F" in use_forces.upper():
				input_param.use_forces = False
			else:
				input_param.use_forces = True

		get_variance, found = read_keyword_argument_same_line("get_variance", lines)
		if found:
			if "F" in get_variance.upper():
				input_param.get_variance = False
			else:
				input_param.get_variance = True

		autograd, found = read_keyword_argument_same_line("autograd", lines)
		if found:
			if "F" in autograd.upper():
				input_param.autograd = False
			else:
				input_param.autograd = True

		train_write, found = read_keyword_argument_same_line("train_write", lines)
		if found:
			if "F" in train_write.upper():
				input_param.train_write = False
			else:
				input_param.train_write = True

		test_write, found = read_keyword_argument_same_line("test_write", lines)
		if found:
			if "F" in test_write.upper():
				input_param.test_write = False
			else:
				input_param.test_write = True

		train_model_save, found = read_keyword_argument_same_line("train_model_save", lines)
		if found:
			if "F" in train_model_save.upper():
				input_param.train_model_save = False
			else:
				input_param.train_model_save = True

		additional_write, found = read_keyword_argument_same_line("additional_write", lines)
		if found:
			if "F" in additional_write.upper():
				input_param.additional_write = False
			else:
				input_param.additional_write = True

		filter, found = read_keyword_argument_same_line("filter", lines)
		if found:
			if "F" in filter.upper():
				input_param.filter = False
			else:
				input_param.filter = True

		# Optional parameters:
		train_file, found = read_keyword_argument_same_line("train_file", lines)
		if found:
			input_param.train_file = glob.glob(train_file)
			input_param.train_file.sort()

		test_file, found = read_keyword_argument_same_line("test_file", lines)
		if found:
			input_param.test_file = glob.glob(test_file)
			input_param.test_file.sort()

		file_format, found = read_keyword_argument_same_line("file_format", lines)
		if found:
			input_param.file_format = str(file_format).lower()

		data_process, found = read_keyword_argument_same_line("data_process", lines)
		if found:
			if 'bat' in data_process.lower():
				input_param.data_process = 'batch'
			elif 'iter' in data_process.lower():
				input_param.data_process = 'iterative'
			else:
				raise Exception("data_process should be either batch or iterative")

		descriptor, found = read_keyword_argument_same_line("descriptor", lines)
		if found:
			if "soap" in descriptor.lower():
				input_param.descriptor = 'soap'
			elif "internal" in descriptor.lower():
				input_param.descriptor = 'internal'
			else:
				input_param.descriptor = 'cartesian coordinates'

		soap_r_cut, found = read_keyword_argument_same_line("soap_r_cut", lines)
		if found:
			input_param.soap_r_cut = float(soap_r_cut)

		soap_n_max, found = read_keyword_argument_same_line("soap_n_max", lines)
		if found:
			input_param.soap_n_max = int(soap_n_max)

		soap_l_max, found = read_keyword_argument_same_line("soap_l_max", lines)
		if found:
			input_param.soap_l_max = int(soap_l_max)

		soap_sigma, found = read_keyword_argument_same_line("soap_sigma", lines)
		if found:
			input_param.soap_sigma = float(soap_sigma)

		soap_rbf, found = read_keyword_argument_same_line("soap_rbf", lines)
		if found:
			if 'poly' in soap_rbf:
				input_param.soap_rbf = 'polynomial'
			else:
				input_param.soap_rbf = 'gto'

		soap_centers, found = read_keyword_list_same_line("soap_centers", lines)
		if found:
			input_param.soap_centers = soap_centers

		soap_n_jobs, found = read_keyword_argument_same_line("soap_n_jobs", lines)
		if found:
			input_param.soap_n_jobs = int(soap_n_jobs)

		scale, found = read_keyword_argument_same_line("scale", lines)
		if found:
			input_param.scale = float(scale)

		weight, found = read_keyword_argument_same_line("weight", lines)
		if found:
			input_param.weight = float(weight)

		noise, found = read_keyword_argument_same_line("noise", lines)
		if found:
			input_param.noise = float(noise)

		noisefactor, found = read_keyword_argument_same_line("noisefactor", lines)
		if found:
			input_param.noisefactor = float(noisefactor)

		batch_size, found = read_keyword_argument_same_line("batch_size", lines)
		if found:
			input_param.train_batch_size = int(batch_size)
			input_param.eval_batch_size = int(batch_size)

		disp_length, found = read_keyword_argument_same_line("disp_length", lines)
		if found:
			input_param.disp_length = float(disp_length)

		num_copy, found = read_keyword_argument_same_line("num_copy", lines)
		if found:
			input_param.num_copy = int(num_copy)

		filter_threshold, found = read_keyword_argument_same_line("filter_threshold", lines)
		if found:
			input_param.filter_threshold = float(filter_threshold)

		return input_param



from .input_parameter import InputParameters
from .inout_process import Train, Test, Augmentation
from .read_input import read_keyword_argument_same_line, read_train_in
from .io_print import *


__all__ = ["input_parameter", "InputParameters",
           "inout_process", "Train", "Test", "Augmentation",
           "read_input", "read_keyword_argument_same_line", "read_train_in",
           "io_print", "io_current_time",
           "io_data_read_finalize", "io_set_descriptor_finalize",
           "io_train_finalize", "io_train_parameters", "io_train_model_save",
           "io_train_model_load", "io_test_evaluation", "io_test_parameters",
           "io_additional_generate_finalize", "io_additional_evaluation", "io_additional_parameters",
           "io_train_write_finalize", "io_test_write_finalize", "io_additional_write_finalize",
           "io_line", "io_print_title", "io_double_line", "io_print_header"]



"""
Collection of utilities for the ASE Tools package.
"""

from subprocess import run, PIPE
import numpy as np
from copy import deepcopy
from ase.io import read, write


__author__ = "Alexander Urban; Jianzhou Qu; In Won Yeu"
__email__ = "aurban@atomistic.net"
__date__ = "2025-07-05"
__version__ = "1.0"


def prepare_neb_images(atoms_init, atoms_final, species=None, pair1=(0, 1), pair2=(0, 2)):
    """
    Prepares two PBC structures for NEB by aligning periodic images and wrapping into the unit cell.

    Parameters
    ----------
    atoms_init : Atoms
        initial structure.
    atoms_final : Atoms
        final structure.
    species : list or set, optional
        Subset of species to use for alignment/wrapping.
    pair1 : tuple, list or set, optional
        (a, b): main direction vector to align with z-axis
    pair2 : tuple, list or set, optional
        (a, c): secondary vector to define the xz-plane
    """

    is_init_pbc = all(atoms_init.pbc)
    is_final_pbc = all(atoms_final.pbc)

    if is_init_pbc and is_final_pbc:
        # Step 1: wrap both structures back into the unit cell
        aligned_init = pbc_wrap(atoms_init, species=species)
        aligned_final = pbc_wrap(atoms_final, species=species)

        # Step 2: Match periodic images
        aligned_final = pbc_match_images(aligned_final, aligned_init)

    elif not is_init_pbc and not is_final_pbc:
        # Non-periodic molecule: align by center
        aligned_init = align_molecule_by_two_vectors(atoms_init, pair1=pair1, pair2=pair2)
        aligned_final = align_molecule_by_two_vectors(atoms_final, pair1=pair1, pair2=pair2)

    else:
        raise ValueError("Both structures must be either periodic or non-periodic.")

    return aligned_init, aligned_final


def pbc_wrap(atoms, species=None, translate=False, eps=1.0e-6):
    """
    Wrap coordinates back into the unit cell.

    Arguments:
      atoms (Atoms): Structure to be wrapped
      species (list or set): Optional list of atomic symbols (e.g., ['O', 'H']) to use for center calculation.
                           If None, use all atoms.
      translate (bool): If True, translate the structure such that the geometric center of all atoms
      (or all selected atoms if species is set) lies exactly in the center of the unit cell.
      After wrapping, the reverse shift will be applied so that the returned structure is compatible with the input structure.
      eps (float): Numerical precision to be used for comparison

    """
    # if not all(atoms.pbc):
    #     return
    # deepcopy is needed to preserve calculators
    new_atoms = deepcopy(atoms)

    if species is None:
        species = set(new_atoms.symbols)
    frac_coords = new_atoms.get_scaled_positions()

    if translate:
        # translate geometric center to the center of the unit cell
        idx = [i for i, s in enumerate(new_atoms.symbols) if s in species]
        C = np.sum(frac_coords[idx], axis=0) / len(idx)
        shift = np.array([0.5, 0.5, 0.5]) - C
    else:
        shift = np.array([0.0, 0.0, 0.0])
    frac_coords += shift

    for iatom, coo in enumerate(frac_coords):
        if new_atoms.symbols[iatom] in species:
            for i in range(3):
                while (coo[i] < 0.0):
                    coo[i] += 1.0
                while (coo[i] >= (1.0 - eps)):
                    coo[i] -= 1.0
    frac_coords -= shift
    new_atoms.set_scaled_positions(frac_coords)

    if hasattr(new_atoms, "calc") and new_atoms._calc is not None:
        new_atoms._calc.atoms.positions[:] = new_atoms.positions[:]
    return new_atoms


def pbc_align(atoms, reference, species=None):
    """
    Return a new Atoms object where each atom is shifted into the same periodic image
    as its counterpart in the reference structure.

    Parameters
    ----------
    atoms : The structure to be aligned.
    reference : The reference structure. Must be the same size and atom order as `atoms`.
    species : list or set, optional
        A list or set of atomic symbols to limit which atoms are aligned.
        If None, all atoms are considered.

    Returns
    -------
    aligned_atoms : Atoms
        A new Atoms object with adjusted atomic positions such that atoms are in
        the same periodic image as in the reference structure.
    """
    # if not all(atoms.pbc):
    #     return
    if len(atoms) != len(reference):
        raise ValueError("Input atoms and reference must contain the same number of atoms.")

    aligned_atoms = deepcopy(atoms)

    if species is not None:
        species = set(species)

    recip = aligned_atoms.cell.reciprocal().T
    cell = np.array(aligned_atoms.cell)

    for i, (sym, pos_ref, pos) in enumerate(zip(reference.symbols, reference.positions, aligned_atoms.positions)):
        if species is not None and sym not in species:
            continue
        # Vector from atoms to reference
        delta = pos_ref - pos
        # Compute how many unit cells to shift
        shift_frac = np.round(np.dot(delta, recip))
        # Convert back to cartesian and apply shift
        shift_cart = np.dot(shift_frac, cell)
        aligned_atoms.positions[i] += shift_cart

    # Update calculator if present
    if aligned_atoms._calc is not None:
        aligned_atoms._calc.atoms.positions[:] = aligned_atoms.positions[:]

    return aligned_atoms


def pbc_match_images(atoms, reference, species=None):
    """
    Adjust atoms' positions such that each selected atom is moved to the periodic image
    closest to its counterpart in the reference structure.

    Parameters
    ----------
    atoms : Atoms
        Structure to be shifted.
    reference : Atoms
        Reference structure (same length and atomic order).
    species : list or set, optional
        Only atoms with matching chemical symbols will be shifted.
        If None, all atoms are considered.

    Returns
    -------
    aligned_atoms : Atoms
        Atoms object with positions adjusted to match reference images.
    """
    if not all(atoms.pbc):
        raise ValueError("pbc_match_images requires full periodic boundary conditions.")

    if len(atoms) != len(reference):
        raise ValueError("atoms and reference must contain the same number of atoms.")

    if species is not None:
        species = set(species)

    aligned_atoms = deepcopy(atoms)
    cell = np.array(atoms.cell)
    recip = atoms.cell.reciprocal().T  # reciprocal lattice vectors

    for i in range(len(atoms)):
        if species is not None and atoms.symbols[i] not in species:
            continue  # skip atoms not in species list

        delta = reference.positions[i] - aligned_atoms.positions[i]
        shift_frac = np.round(np.dot(delta, recip))  # periodic shift in fractional coordinates
        shift_cart = np.dot(shift_frac, cell)  # convert back to Cartesian
        aligned_atoms.positions[i] += shift_cart

    # Update attached calculator if necessary
    if aligned_atoms._calc is not None:
        aligned_atoms._calc.atoms.positions[:] = aligned_atoms.positions[:]

    return aligned_atoms


def align_molecule_by_two_vectors(atoms_target, pair1, pair2):
    """
    Align atoms_target using two direction vectors:
    - pair1 = (a, b): main direction vector to align with z-axis
    - pair2 = (a, c): secondary vector to define the xz-plane
    Also moves atom a to origin.
    """
    a, b = pair1
    _, c = pair2

    new_atoms = deepcopy(atoms_target)
    pos = new_atoms.get_positions()

    # Compute local frame in target
    v1 = pos[b] - pos[a]  # main direction (to z)
    v2 = pos[c] - pos[a]  # secondary direction

    z_axis = v1 / np.linalg.norm(v1)
    x_temp = v2 - np.dot(v2, z_axis) * z_axis  # project v2 to plane orthogonal to z
    x_axis = x_temp / np.linalg.norm(x_temp)
    y_axis = np.cross(z_axis, x_axis)

    R = np.vstack([x_axis, y_axis, z_axis]).T  # rotation matrix from local to global frame

    # Apply inverse rotation to align local frame to global axes
    rotated = (pos - pos[a]) @ R
    new_atoms.set_positions(rotated)

    return new_atoms


def align_molecule_centers(atoms_ref, atoms_target, species=None):
    """
    Translate `atoms_target` such that its center of geometry matches that of `atoms_ref`.

    Parameters:
    -----------
    atoms_ref (Atoms): Reference structure
    atoms_target (Atoms): Structure to be shifted
    species (list or set): Optional list of atomic symbols (e.g., ['O', 'H']) to use for center calculation.
                           If None, use all atoms.

    Returns:
    --------
    new_atoms (Atoms): Shifted copy of `atoms_target` aligned to `atoms_ref`
    """

    new_atoms = deepcopy(atoms_target)

    if species is None:
        idx_ref = np.arange(len(atoms_ref))
        idx_target = np.arange(len(new_atoms))
    else:
        idx_ref = [i for i, s in enumerate(atoms_ref.get_chemical_symbols()) if s in species]
        idx_target = [i for i, s in enumerate(new_atoms.get_chemical_symbols()) if s in species]

    # Compute centers
    center_ref = atoms_ref.get_positions()[idx_ref].mean(axis=0)
    center_target = new_atoms.get_positions()[idx_target].mean(axis=0)

    shift = center_ref - center_target
    new_atoms.translate(shift)

    if hasattr(new_atoms, "calc") and new_atoms._calc is not None:
        new_atoms._calc.atoms.positions[:] = new_atoms.positions[:]

    return new_atoms


def runcmd(command, maxtime=300):
    """
    Run cmd command based on `bash`.
    Args:
        command: String
            Can be the same way in command line.
        maxtime: Int
            The maximum time (second) for running this command.
    """
    runit = run(
        command, stdout=PIPE, stderr=PIPE, shell=True, encoding="utf-8",
        executable="/bin/bash", timeout=maxtime
    )
    if runit.returncode == 0:
        print("Succeed to run command", command)
    else:
        print("Error:", runit)
    return None


def csv2list(csvlist):
    """
    Convert list of comma-separated integers to an actual list.

    Ranges are defined by ":".  For example:

       [1, "2", "3:7", "8,9:11"]

    will be converted to

       [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]

    """

    def expand_range(r):
        if ":" not in r:
            return [int(r)]
        i0, i1 = [int(s) for s in r.split(":")]
        return list(range(i0, i1 + 1))

    lst = []
    for item in csvlist:
        try:
            lst.append([int(item)])
        except ValueError:
            sublst = []
            for item2 in item.split(","):
                sublst += expand_range(item2)
            lst.append(sublst)
    return lst


def pbc_group(atoms, atom_groups):
    """
    Select periodic images of grouped atoms such that all grouped atoms
    are closest to each other.

    Arguments:
      atoms (Atoms): Structure to be aligned
      atom_groups (list of lists): Lists with grouped atoms
    """
    if not all(atoms.pbc):
        return

    def group(at, ref):
        """ Select periodic image of 'at' closest to 'ref'. """
        coo = atoms.positions[at]
        coo_ref = atoms.positions[ref]
        vec = coo_ref - coo
        vec = np.round(np.dot(vec, atoms.cell.reciprocal().T))
        vec = np.dot(vec, np.array(atoms.cell))
        return coo + vec

    for grp in atom_groups:
        for i in grp[1:]:
            atoms.positions[i] = group(i, grp[0])
    if atoms._calc is not None:
        atoms._calc.atoms.positions[:] = atoms.positions[:]
    return atoms



from .reference_data import ReferenceData
from .reference_data_internal import ReferenceDataInternal
from .additional_data import AdditionalData
from .additional_data_internal import AdditionalDataInternal
from .prepare_data import get_N_batch, get_batch_indexes_N_batch, standard_output, inverse_standard_output

__all__ = ["reference_data", "additional_data",
           "reference_data_internal", "ReferenceDataInternal",
           "ReferenceData", "AdditionalData",
           "additional_data_internal", "AdditionalDataInternal",
           "prepare_data", "get_N_batch", "get_batch_indexes_N_batch", "standard_output", "inverse_standard_output"]



import sys
import torch

from aenet_gpr.inout import Train, Test, Augmentation
from aenet_gpr.inout.read_input import read_train_in
from aenet_gpr.inout.io_print import *


def main():
    device = "cpu"
    if torch.cuda.is_available():
        device = "cuda:0"

    io_print_header()

    # 1. Read train input
    if len(sys.argv) > 1:
        input_file = sys.argv[1]
    else:
        input_file = "train.in"
    input_param = read_train_in(input_file)
    input_param.update_soap_param()
    input_param.device = device

    # if input_param.verbose:
    #     io_input_reading(input_param)

    if len(input_param.train_file) != 0:
        io_print_title(text='Train')

        train_process = Train(input_param=input_param)
        train_process.read_reference_train_data()
        train_process.train_model()
        if input_param.train_model_save:
            train_process.train_model_save()
        if input_param.train_write:
            train_process.write_reference_train_xsf()

    if len(input_param.test_file) != 0:
        io_print_title(text='Test')

        test_process = Test(input_param=input_param)
        if len(input_param.train_file) != 0:
            test_process.load_train_model(train_data=train_process.train_data)
        else:
            test_process.load_train_model(train_data=None)
        test_process.read_reference_test_data()
        test_process.model_test_evaluation()
        if input_param.test_write:
            test_process.write_reference_test_xsf()

    if input_param.additional_write:
        io_print_title(text='Augmentation')

        augmentation_process = Augmentation(input_param=input_param)
        if len(input_param.train_file) != 0:
            augmentation_process.load_train_model(train_data=train_process.train_data)
        else:
            augmentation_process.load_train_model(train_data=None)
        augmentation_process.generate_additional_structures()
        augmentation_process.model_additional_evaluation()
        augmentation_process.write_additional_xsf()


if __name__ == "__main__":
    main()